# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Classes defining OT problem(s) (objective function + utilities)."""

from typing import TYPE_CHECKING, Optional, Tuple, Union

import jax
import jax.numpy as jnp
from typing_extensions import Literal

from ott.geometry import epsilon_scheduler, geometry, low_rank, pointcloud
from ott.problems.linear import linear_problem
from ott.problems.quadratic import quadratic_costs
from ott.types import Transport

if TYPE_CHECKING:
  from ott.solvers.linear import sinkhorn_lr

__all__ = ["QuadraticProblem"]


@jax.tree_util.register_pytree_node_class
class QuadraticProblem:
  r"""Quadratic regularized OT problem.

  The quadratic loss of a single OT matrix is assumed to
  have the form given in :cite:`peyre:16`, eq. 4.

  The two geometries below parameterize matrices :math:`C` and :math:`\bar{C}`
  in that equation. The function :math:`L` (of two real values) in that equation
  is assumed to match the form given in eq. 5., with our notations:

  .. math::

    L(x, y) = lin1(x) + lin2(y) - quad1(x) * quad2(y)

  Args:
    geom_xx: Ground geometry of the first space.
    geom_yy: Ground geometry of the second space.
    geom_xy: Geometry defining the linear penalty term for Fused Gromov
      Wasserstein. If `None`, the problem reduces to a
      plain Gromov Wasserstein problem.
    fused_penalty: multiplier of the linear term in Fused Gromov Wasserstein,
      i.e. problem = purely quadratic + fused_penalty * linear problem.
      Ignored if ``geom_xy`` is not specified.
    scale_cost: option to rescale the cost matrices:

      - if `True`, use the default for each geometry.
      - if `False`, keep the original scaling in geometries.
      - if :class:`str`, use a specific method available in
        :class:`~ott.geometry.geometry.Geometry` or
        :class:`~ott.geometry.pointcloud.PointCloud`.
      - if `None`, do not scale the cost matrices.

    a: jnp.ndarray[n] representing the probability weights of the samples
      from geom_xx. If None, it will be uniform.
    b: jnp.ndarray[n] representing the probability weights of the samples
      from geom_yy. If None, it will be uniform.
    loss: a 2-tuple of 2-tuples of Callable. The first tuple is the linear
      part of the loss (see in the pydoc of the class lin1, lin2). The second
      one is the quadratic part (quad1, quad2). By default, the loss
      is set as the 4 functions representing the squared Euclidean loss, and
      this property is taken advantage of in subsequent computations. See
      Alternatively, KL loss can be specified in no less optimized way.
    tau_a: if lower that 1.0, defines how much unbalanced the problem is on
      the first marginal.
    tau_b: if lower that 1.0, defines how much unbalanced the problem is on
      the second marginal.
    gw_unbalanced_correction: Whether the unbalanced version of
      :cite:`sejourne:21` is used. Otherwise ``tau_a`` and ``tau_b`` only affect
      the inner Sinkhorn loop.
    ranks: Ranks of the cost matrices, see
      :meth:`~ott.geometry.geometry.Geometry.to_LRCGeometry`. Used when
      geometries are *not* :class:`~ott.geometry.pointcloud.PointCloud` with
      `'sqeucl'` cost function. If `-1`, the geometries will not be converted
      to low-rank. If :class:`tuple`, it specifies the ranks of ``geom_xx``,
      ``geom_yy`` and ``geom_xy``, respectively. If :class:`int`, rank is shared
      across all geometries.
    tolerances: Tolerances used when converting geometries to low-rank. Used
      when geometries are not :class:`~ott.geometry.pointcloud.PointCloud` with
      `'sqeucl'` cost. If :class:`float`, it is shared across all geometries.
  """

  def __init__(
      self,
      geom_xx: geometry.Geometry,
      geom_yy: geometry.Geometry,
      geom_xy: Optional[geometry.Geometry] = None,
      fused_penalty: float = 1.0,
      scale_cost: Optional[Union[bool, float, str]] = False,
      a: Optional[jnp.ndarray] = None,
      b: Optional[jnp.ndarray] = None,
      loss: Union[Literal['sqeucl', 'kl'], quadratic_costs.GWLoss] = 'sqeucl',
      tau_a: Optional[float] = 1.0,
      tau_b: Optional[float] = 1.0,
      gw_unbalanced_correction: bool = True,
      ranks: Union[int, Tuple[int, ...]] = -1,
      tolerances: Union[float, Tuple[float, ...]] = 1e-2,
  ):
    assert fused_penalty > 0, fused_penalty
    self._geom_xx = geom_xx._set_scale_cost(scale_cost)
    self._geom_yy = geom_yy._set_scale_cost(scale_cost)
    self._geom_xy = (
        None if geom_xy is None else geom_xy._set_scale_cost(scale_cost)
    )
    self.fused_penalty = fused_penalty
    self.scale_cost = scale_cost
    self._a = a
    self._b = b
    self.tau_a = tau_a
    self.tau_b = tau_b
    self.gw_unbalanced_correction = gw_unbalanced_correction
    self.ranks = ranks
    self.tolerances = tolerances

    self._loss_name = loss
    if self._loss_name == 'sqeucl':
      self.loss = quadratic_costs.make_square_loss()
    elif loss == 'kl':
      self.loss = quadratic_costs.make_kl_loss()
    else:
      self.loss = loss

  def marginal_dependent_cost(
      self, marginal_1: jnp.ndarray, marginal_2: jnp.ndarray
  ) -> low_rank.LRCGeometry:
    r"""Initialise cost term that depends on the marginals of the transport.

    Uses the first term in eq. 6, p. 1 of :cite:`peyre:16`.

    Let :math:`p` [num_a,] be the marginal of the transport matrix for samples
    from `geom_xx` and :math:`q` [num_b,] be the marginal of the transport
    matrix for samples from `geom_yy`. `cost_xx` (resp. `cost_yy`) is the
    cost matrix of `geom_xx` (resp. `geom_yy`). The cost term that
    depends on these marginals can be written as:

    `marginal_dep_term` = `lin1`(`cost_xx`) :math:`p \mathbb{1}_{num_b}^T`
                      + (`lin2`(`cost_yy`) :math:`q \mathbb{1}_{num_a}^T)^T`

    Args:
      marginal_1: jnp.ndarray<float>[num_a,], marginal of the transport matrix
       for samples from geom_xx
      marginal_2: jnp.ndarray<float>[num_b,], marginal of the transport matrix
       for samples from geom_yy

    Returns:
      Low-rank geometry.
    """
    if self._loss_name == 'sqeucl':  # quadratic apply, efficient for LR
      tmp1 = self.geom_xx.apply_square_cost(marginal_1, axis=1)
      tmp2 = self.geom_yy.apply_square_cost(marginal_2, axis=1)
    else:
      f1, f2 = self.linear_loss
      tmp1 = apply_cost(self.geom_xx, marginal_1, axis=1, fn=f1)
      tmp2 = apply_cost(self.geom_yy, marginal_2, axis=1, fn=f2)
    x_term = jnp.concatenate((tmp1, jnp.ones_like(tmp1)), axis=1)
    y_term = jnp.concatenate((jnp.ones_like(tmp2), tmp2), axis=1)
    return low_rank.LRCGeometry(cost_1=x_term, cost_2=y_term)

  def cost_unbalanced_correction(
      self,
      transport_matrix: jnp.ndarray,
      marginal_1: jnp.ndarray,
      marginal_2: jnp.ndarray,
      epsilon: float,
  ) -> float:
    r"""Calculate cost term from the quadratic divergence when unbalanced.

    In the unbalanced setting (``tau_a < 1.0 or tau_b < 1.0``), the
    introduction of a quadratic divergence :cite:`sejourne:21` adds a term
    to the GW local cost.

    Let :math:`a` [num_a,] be the target weights for samples
    from geom_xx and :math:`b` [num_b,] be the target weights
    for samples from `geom_yy`. Let :math:`P` [num_a, num_b] be the transport
    matrix, :math:`P1` the first marginal and :math:`P^T1` the second marginal.
    The term of the cost matrix coming from the quadratic KL in the
    unbalanced case can be written as:

    `unbalanced_correction_term` =
        :math:`tau_a / (1 - tau_a) * \sum(KL(P1|a))`
        :math:`+ tau_b / (1 - tau_b) * \sum(KL(P^T1|b))`
        :math:`+ epsilon * \sum(KL(P|ab'))`

    Args:
      transport_matrix: jnp.ndarray<float>[num_a, num_b], transport matrix.
      marginal_1: jnp.ndarray<float>[num_a,], marginal of the transport matrix
        for samples from :attr:`geom_xx`.
      marginal_2: jnp.ndarray<float>[num_b,], marginal of the transport matrix
        for samples from :attr:`geom_yy`.
      epsilon: regulariser.
      delta: small quantity to avoid diverging KLs.

    Returns:
      The cost term.
    """

    def regulariser(tau: float) -> float:
      return epsilon._target_init * tau / (1.0 - tau) if tau != 1.0 else 0

    marginal_1loga = jax.scipy.special.xlogy(marginal_1, self.a).sum()
    marginal_2logb = jax.scipy.special.xlogy(marginal_2, self.b).sum()
    cost = regulariser(
        self.tau_a
    ) * (-jax.scipy.special.entr(marginal_1).sum() - marginal_1loga)
    cost += regulariser(
        self.tau_b
    ) * (-jax.scipy.special.entr(marginal_2).sum() - marginal_2logb)
    cost += epsilon._target_init * jax.scipy.special.xlogy(
        transport_matrix, transport_matrix
    ).sum()
    return cost

  def init_transport(self) -> jnp.ndarray:
    """Initialise the transport matrix."""
    # TODO(oliviert, cuturi): consider passing a custom initialization.
    a = jax.lax.stop_gradient(self.a)
    b = jax.lax.stop_gradient(self.b)
    return a[:, None] * b[None, :]

  def init_transport_mass(self) -> float:
    """Initialise the transport mass.

    Returns:
      The sum of the elements of the normalised transport matrix.
    """
    a = jax.lax.stop_gradient(self.a)
    b = jax.lax.stop_gradient(self.b)
    return a.sum() * b.sum()

  def update_lr_geom(
      self, lr_sink: 'sinkhorn_lr.LRSinkhornOutput'
  ) -> geometry.Geometry:
    """Recompute (possibly LRC) linearization using LR Sinkhorn output."""
    marginal_1 = lr_sink.marginal(1)
    marginal_2 = lr_sink.marginal(0)
    marginal_cost = self.marginal_dependent_cost(marginal_1, marginal_2)

    # Extract factors from LR Sinkhorn output
    q, r, inv_sqg = lr_sink.q, lr_sink.r, 1.0 / jnp.sqrt(lr_sink.g)
    # Distribute middle marginal evenly across both factors.
    q, r = q * inv_sqg[None, :], r * inv_sqg[None, :]

    # Handle LRC Geometry case.
    h1, h2 = self.quad_loss
    tmp1 = apply_cost(self.geom_xx, q, axis=1, fn=h1)
    tmp2 = apply_cost(self.geom_yy, r, axis=1, fn=h2)
    if self.is_low_rank:
      geom = low_rank.LRCGeometry(cost_1=tmp1, cost_2=-tmp2) + marginal_cost
      if self.is_fused:
        geom = geom + self.geom_xy
    else:
      cost_matrix = marginal_cost.cost_matrix - jnp.dot(tmp1, tmp2.T)
      cost_matrix += self.fused_penalty * self._fused_cost_matrix
      geom = geometry.Geometry(cost_matrix=cost_matrix)
    return geom

  def update_linearization(
      self,
      transport: Transport,
      epsilon: Optional[Union[epsilon_scheduler.Epsilon, float]] = None,
      old_transport_mass: float = 1.0
  ) -> linear_problem.LinearProblem:
    """Update linearization of GW problem by updating cost matrix.

    If the problem is balanced (``tau_a = 1.0 and tau_b = 1.0``), the equation
    follows eq. 6, p. 1 of :cite:`peyre:16`.

    If the problem is unbalanced (``tau_a < 1.0 or tau_b < 1.0``), two cases are
    possible, as explained in :meth:`init_linearization` above.
    Finally, it is also possible to consider a Fused Gromov Wasserstein problem.
    Details about the resulting cost matrix are also given in
    :meth:`init_linearization`.

    Args:
      transport: Solution of the linearization of the quadratic problem.
      epsilon: An epsilon scheduler or a float passed on to the linearization.
      old_transport_mass: Sum of the elements of the transport matrix at the
        previous iteration.

    Returns:
      Updated linear OT problem, a new local linearization of GW problem.
    """
    rescale_factor = 1.0
    unbalanced_correction = 0.0

    if not self.is_balanced:
      marginal_1 = transport.marginal(axis=1)
      transport_mass = jax.lax.stop_gradient(marginal_1.sum())
      rescale_factor = jnp.sqrt(old_transport_mass / transport_mass)

    marginal_1 = transport.marginal(axis=1) * rescale_factor
    marginal_2 = transport.marginal(axis=0) * rescale_factor
    marginal_cost = self.marginal_dependent_cost(marginal_1, marginal_2)

    transport_matrix = transport.matrix * rescale_factor

    if not self.is_balanced:
      # Rescales transport for Unbalanced GW according to Sejourne et al (2021).
      transport_mass = jax.lax.stop_gradient(marginal_1.sum())
      epsilon = update_epsilon_unbalanced(epsilon, transport_mass)
      unbalanced_correction = self.cost_unbalanced_correction(
          transport_matrix, marginal_1, marginal_2, epsilon
      )

    h1, h2 = self.quad_loss
    tmp = apply_cost(self.geom_xx, transport_matrix, axis=1, fn=h1)
    tmp = apply_cost(self.geom_yy, tmp.T, axis=1, fn=h2).T

    cost_matrix = marginal_cost.cost_matrix - tmp + unbalanced_correction
    cost_matrix += self.fused_penalty * self._fused_cost_matrix * rescale_factor

    geom = geometry.Geometry(cost_matrix=cost_matrix, epsilon=epsilon)
    return linear_problem.LinearProblem(
        geom, self.a, self.b, tau_a=self.tau_a, tau_b=self.tau_b
    )

  def update_lr_linearization(
      self, lr_sink: 'sinkhorn_lr.LRSinkhornOutput'
  ) -> linear_problem.LinearProblem:
    """Update a Quad problem linearization using a LR Sinkhorn."""
    return linear_problem.LinearProblem(
        self.update_lr_geom(lr_sink),
        self.a,
        self.b,
        tau_a=self.tau_a,
        tau_b=self.tau_b
    )

  @property
  def _fused_cost_matrix(self) -> Union[float, jnp.ndarray]:
    if not self.is_fused:
      return 0.
    if isinstance(
        self.geom_xy, pointcloud.PointCloud
    ) and self.geom_xy.is_online:
      return self.geom_xy._compute_cost_matrix() * self.geom_xy.inv_scale_cost
    return self.geom_xy.cost_matrix

  @property
  def _is_low_rank_convertible(self) -> bool:

    def convertible(geom: geometry.Geometry) -> bool:
      return isinstance(geom, low_rank.LRCGeometry) or (
          isinstance(geom, pointcloud.PointCloud) and geom.is_squared_euclidean
      )

    if self.is_low_rank:
      return True

    geom_xx, geom_yy, geom_xy = self.geom_xx, self.geom_yy, self.geom_xy
    # either explicitly via cost factorization or implicitly (e.g., a PC)
    return self.ranks != -1 or (
        convertible(geom_xx) and convertible(geom_yy) and
        (geom_xy is None or convertible(geom_xy))
    )

  def to_low_rank(self, seed: int = 0) -> "QuadraticProblem":
    """Convert geometries to low-rank.

    Args:
      seed: Random seed.

    Returns:
      Quadratic problem with low-rank geometries.
    """

    def convert(
        vals: Union[int, float, Tuple[Union[int, float], ...]]
    ) -> Tuple[Union[int, float], ...]:
      size = 2 + self.is_fused
      if isinstance(vals, (int, float)):
        return (vals,) * 3
      assert len(vals) == size, vals
      return vals + (None,) * (3 - size)

    if self.is_low_rank:
      return self

    (geom_xx, geom_yy, geom_xy, *children), aux_data = self.tree_flatten()
    (s1, s2, s3) = jax.random.split(jax.random.PRNGKey(seed), 3)[:, 0]
    (r1, r2, r3), (t1, t2, t3) = convert(self.ranks), convert(self.tolerances)

    geom_xx = geom_xx.to_LRCGeometry(rank=r1, tol=t1, seed=s1)
    geom_yy = geom_yy.to_LRCGeometry(rank=r2, tol=t2, seed=s2)
    if self.is_fused:
      if isinstance(
          geom_xy, pointcloud.PointCloud
      ) and geom_xy.is_squared_euclidean:
        geom_xy = geom_xy.to_LRCGeometry(scale=self.fused_penalty)
      else:
        geom_xy = geom_xy.to_LRCGeometry(
            rank=r3, tol=t3, seed=s3, scale=self.fused_penalty
        )

    return type(self).tree_unflatten(
        aux_data, [geom_xx, geom_yy, geom_xy] + children
    )

  @property
  def geom_xx(self) -> geometry.Geometry:
    """Geometry of the first space."""
    return self._geom_xx

  @property
  def geom_yy(self) -> geometry.Geometry:
    """Geometry of the second space."""
    return self._geom_yy

  @property
  def geom_xy(self) -> Optional[geometry.Geometry]:
    """Geometry of the joint space."""
    return self._geom_xy

  @property
  def a(self) -> jnp.ndarray:
    """First marginal."""
    num_a = self.geom_xx.shape[0]
    return jnp.ones((num_a,)) / num_a if self._a is None else self._a

  @property
  def b(self) -> jnp.ndarray:
    """Second marginal."""
    num_b = self.geom_yy.shape[0]
    return jnp.ones((num_b,)) / num_b if self._b is None else self._b

  @property
  def is_fused(self) -> bool:
    """Whether the problem is fused."""
    return self.geom_xy is not None

  @property
  def is_low_rank(self) -> bool:
    """Whether all geometries are low-rank."""
    return (
        isinstance(self.geom_xx, low_rank.LRCGeometry) and
        isinstance(self.geom_yy, low_rank.LRCGeometry) and
        (not self.is_fused or isinstance(self.geom_xy, low_rank.LRCGeometry))
    )

  @property
  def linear_loss(self) -> Tuple[quadratic_costs.Loss, quadratic_costs.Loss]:
    """Linear part of the Gromov-Wasserstein loss."""
    return self.loss.f1, self.loss.f2

  @property
  def quad_loss(self) -> Tuple[quadratic_costs.Loss, quadratic_costs.Loss]:
    """Quadratic part of the Gromov-Wasserstein loss."""
    return self.loss.h1, self.loss.h2

  @property
  def is_balanced(self) -> bool:
    """Whether the problem is balanced."""
    return ((not self.gw_unbalanced_correction) or
            (self.tau_a == 1.0 and self.tau_b == 1.0))

  def tree_flatten(self):
    return ([self.geom_xx, self.geom_yy, self.geom_xy, self._a, self._b], {
        'tau_a': self.tau_a,
        'tau_b': self.tau_b,
        'loss': self._loss_name,
        'fused_penalty': self.fused_penalty,
        'scale_cost': self.scale_cost,
        'gw_unbalanced_correction': self.gw_unbalanced_correction,
        'ranks': self.ranks,
        'tolerances': self.tolerances
    })

  @classmethod
  def tree_unflatten(cls, aux_data, children):
    geoms, (a, b) = children[:3], children[3:]
    return cls(*geoms, a=a, b=b, **aux_data)


def update_epsilon_unbalanced(
    epsilon: Union[float, epsilon_scheduler.Epsilon], transport_mass: float
) -> epsilon_scheduler.Epsilon:
  updated_epsilon = epsilon_scheduler.Epsilon.make(epsilon)
  updated_epsilon._scale_epsilon = (
      updated_epsilon._scale_epsilon * transport_mass
  )
  return updated_epsilon


def apply_cost(
    geom: geometry.Geometry, arr: jnp.ndarray, *, axis: int,
    fn: quadratic_costs.Loss
) -> jnp.ndarray:
  return geom.apply_cost(arr, axis=axis, fn=fn.func, is_linear=fn.is_linear)



# Copyright 2022 Apple.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""A Jax version of the W barycenter algorithm (Cuturi Doucet 2014)."""
import functools
from typing import Any, NamedTuple, Optional, Tuple

import jax
import jax.numpy as jnp

from ott.geometry import pointcloud
from ott.math import fixed_point_loop
from ott.math import utils as mu
from ott.problems.linear import barycenter_problem, linear_problem
from ott.solvers import was_solver

__all__ = ["BarycenterState", "WassersteinBarycenter"]


class BarycenterState(NamedTuple):
  """Holds the state of the Wasserstein barycenter solver.

  Args:
    costs: Holds the sequence of regularized GW costs seen through the outer
      loop of the solver.
    linear_convergence: Holds the sequence of bool convergence flags of the
      inner Sinkhorn iterations.
    errors: Holds sequence of vectors of errors of the Sinkhorn algorithm
      at each iteration.
    x: barycenter points.
    a: barycenter weights.
  """

  costs: Optional[jnp.ndarray] = None
  linear_convergence: Optional[jnp.ndarray] = None
  errors: Optional[jnp.ndarray] = None
  x: Optional[jnp.ndarray] = None
  a: Optional[jnp.ndarray] = None

  def set(self, **kwargs: Any) -> 'BarycenterState':
    """Return a copy of self, possibly with overwrites."""
    return self._replace(**kwargs)

  def update(
      self, iteration: int, bar_prob: barycenter_problem.BarycenterProblem,
      linear_ot_solver: Any, store_errors: bool
  ) -> 'BarycenterState':
    seg_y, seg_b = bar_prob.segmented_y_b

    @functools.partial(jax.vmap, in_axes=[None, None, 0, 0])
    def solve_linear_ot(
        a: Optional[jnp.ndarray], x: jnp.ndarray, b: jnp.ndarray, y: jnp.ndarray
    ):
      out = linear_ot_solver(
          linear_problem.LinearProblem(
              pointcloud.PointCloud(
                  x,
                  y,
                  src_mask=a > 0.,
                  tgt_mask=b > 0.,
                  cost_fn=bar_prob.cost_fn,
                  epsilon=bar_prob.epsilon
              ), a, b
          )
      )
      return (
          out.reg_ot_cost, out.converged, out.matrix,
          out.errors if store_errors else None
      )

    if bar_prob.debiased:
      raise NotImplementedError(
          "Debiased version of continuous Wasserstein barycenter "
          "not yet implemented."
      )

    reg_ot_costs, convergeds, matrices, errors = solve_linear_ot(
        self.a, self.x, seg_b, seg_y
    )

    cost = jnp.sum(reg_ot_costs * bar_prob.weights)
    updated_costs = self.costs.at[iteration].set(cost)
    converged = jnp.all(convergeds)
    linear_convergence = self.linear_convergence.at[iteration].set(converged)

    if store_errors and self.errors is not None:
      errors = self.errors.at[iteration, :, :].set(errors)
    else:
      errors = None

    # Approximation of barycenter as barycenter of barycenters per measure.

    barycenters_per_measure = mu.barycentric_projection(
        matrices, seg_y, bar_prob.cost_fn
    )

    x_new = jax.vmap(
        bar_prob.cost_fn.barycenter, in_axes=[None, 1]
    )(bar_prob.weights, barycenters_per_measure)

    return self.set(
        costs=updated_costs,
        linear_convergence=linear_convergence,
        errors=errors,
        x=x_new
    )


@jax.tree_util.register_pytree_node_class
class WassersteinBarycenter(was_solver.WassersteinSolver):
  """A Continuous Wasserstein barycenter solver, built on generic template."""

  def __call__(
      self,
      bar_prob: barycenter_problem.BarycenterProblem,
      bar_size: int = 100,
      x_init: Optional[jnp.ndarray] = None,
      rng: int = 0
  ) -> BarycenterState:
    bar_fn = jax.jit(iterations, static_argnums=1) if self.jit else iterations
    out = bar_fn(self, bar_size, bar_prob, x_init, rng)
    return out

  def init_state(
      self,
      bar_prob: barycenter_problem.BarycenterProblem,
      bar_size: int,
      x_init: Optional[jnp.ndarray] = None,
      # TODO(michalk8): change the API to pass the PRNG key directly
      rng: int = 0,
  ) -> BarycenterState:
    """Initialize the state of the Wasserstein barycenter iterations.

    Args:
      bar_prob: The barycenter problem.
      bar_size: Size of the barycenter.
      x_init: Initial barycenter estimate of shape ``[bar_size, ndim]``.
        If `None`, ``bar_size`` points will be sampled from the input
        measures according to their weights
        :attr:`~ott.problems.linear.barycenter_problem.BarycenterProblem.flattened_y`.
      rng: Seed for :func:`jax.random.PRNGKey`.

    Returns:
      The initial barycenter state.
    """
    if x_init is not None:
      assert bar_size == x_init.shape[0]
      x = x_init
    else:
      # sample randomly points in the support of the y measures
      indices_subset = jax.random.choice(
          jax.random.PRNGKey(rng),
          a=bar_prob.flattened_y.shape[0],
          shape=(bar_size,),
          replace=False,
          p=bar_prob.flattened_b
      )
      x = bar_prob.flattened_y[indices_subset, :]

    # TODO(cuturi) expand to non-uniform weights for barycenter.
    a = jnp.ones((bar_size,)) / bar_size
    num_iter = self.max_iterations
    if self.store_inner_errors:
      errors = -jnp.ones((
          num_iter, bar_prob.num_measures,
          self.linear_ot_solver.outer_iterations
      ))
    else:
      errors = None
    return BarycenterState(
        -jnp.ones((num_iter,)), -jnp.ones((num_iter,)), errors, x, a
    )

  def output_from_state(self, state: BarycenterState) -> BarycenterState:
    # TODO(michalk8): create an output variable to match rest of the framework
    return state


def iterations(
    solver: WassersteinBarycenter, bar_size: int,
    bar_prob: barycenter_problem.BarycenterProblem, x_init: jnp.ndarray,
    rng: int
) -> BarycenterState:
  """Jittable Wasserstein barycenter outer loop."""

  def cond_fn(
      iteration: int, constants: Tuple[WassersteinBarycenter,
                                       barycenter_problem.BarycenterProblem],
      state: BarycenterState
  ) -> bool:
    solver, _ = constants
    return solver._continue(state, iteration)

  def body_fn(
      iteration, constants: Tuple[WassersteinBarycenter,
                                  barycenter_problem.BarycenterProblem],
      state: BarycenterState, compute_error: bool
  ) -> BarycenterState:
    del compute_error  # Always assumed True
    solver, bar_prob = constants
    return state.update(
        iteration, bar_prob, solver.linear_ot_solver, solver.store_inner_errors
    )

  state = fixed_point_loop.fixpoint_iter(
      cond_fn=cond_fn,
      body_fn=body_fn,
      min_iterations=solver.min_iterations,
      max_iterations=solver.max_iterations,
      inner_iterations=1,
      constants=(solver, bar_prob),
      state=solver.init_state(bar_prob, bar_size, x_init, rng)
  )

  return solver.output_from_state(state)



import abc
from typing import (
    Any,
    Dict,
    Generic,
    Hashable,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

import jax
import jax.experimental.host_callback as hcb
import jax.experimental.sparse as jesp
import jax.numpy as jnp
import jax.scipy as jsp
import numpy as np
import scipy.sparse as sp

try:
  from sksparse import cholmod
except ImportError:
  cholmod = None

__all__ = ["DenseCholeskySolver", "SparseCholeskySolver"]

T = TypeVar("T")


@jax.tree_util.register_pytree_node_class
class CholeskySolver(abc.ABC, Generic[T]):
  """Base class for Cholesky linear solver.

  Args:
    A: Symmetric positive definite matrix of shape ``[n, n]``.
  """

  def __init__(self, A: T):
    self._A = A
    self._L: Optional[T] = None  # Cholesky factor

  def solve(self, b: jnp.ndarray) -> jnp.ndarray:
    """Solve the linear system :math:`A * x = b`.

    Args:
        b: Vector of shape ``[n,]``.

    Returns:
        The solution of shape ``[n,]``.
    """
    return self._solve(self.L, b)

  @abc.abstractmethod
  def _decompose(self, A: T) -> Optional[T]:
    """Decompose matrix ``A`` into Cholesky factor."""

  @abc.abstractmethod
  def _solve(self, L: Optional[T], b: jnp.ndarray) -> jnp.ndarray:
    """Solve a triangular linear system :math:`L * x = b`."""

  @classmethod
  def create(cls, A: Union[T, sp.spmatrix], **kwargs: Any) -> "CholeskySolver":
    """Instantiate sparse or dense Cholesky solver.

    And optionally convert :class:`scipy.sparse.spmatrix` to its
    :mod:`jax` equivalent.

    Args:
      A: Symmetric positive definite matrix of shape ``[n, n]``.
      kwargs: Keyword arguments for the solver initialization.

    Returns:
      Sparse or dense Cholesky solver.
    """
    if isinstance(A, sp.spmatrix):
      A = _scipy_sparse_to_jax(A)
    if isinstance(A, (jesp.CSR, jesp.CSC, jesp.BCOO)):
      return SparseCholeskySolver(A, **kwargs)
    return DenseCholeskySolver(A, **kwargs)

  @property
  def A(self) -> jnp.ndarray:
    """Symmetric positive definite matrix of shape ``[n, n]``."""
    return self._A

  @property
  def L(self) -> Optional[T]:
    """Cholesky factor of :attr:`A`."""
    if self._L is None:
      self._L = self._decompose(self.A)
    return self._L

  def tree_flatten(self) -> Tuple[Sequence[Any], Dict[str, Any]]:
    return (self.A, self.L), {}

  @classmethod
  def tree_unflatten(
      cls, aux_data: Mapping[str, Any], children: Sequence[Any]
  ) -> "CholeskySolver":
    A, L = children
    obj = cls(A, **aux_data)
    obj._L = L
    return obj


@jax.tree_util.register_pytree_node_class
class DenseCholeskySolver(CholeskySolver[jnp.ndarray]):
  """Dense Cholesky solver.

  Args:
    A: Symmetric positive definite matrix of shape ``[n, n]``.
    lower: Whether to compute lower-triangular Cholesky factor.
    kwargs: Additional keyword arguments, currently ignored.
  """

  def __init__(self, A: T, lower: bool = True, **kwargs: Any):
    del kwargs
    super().__init__(A)
    self._lower = lower

  def _decompose(self, A: T) -> Optional[T]:
    return jsp.linalg.cholesky(A, lower=self._lower)

  def _solve(self, L: Optional[T], b: jnp.ndarray) -> jnp.ndarray:
    return jsp.linalg.solve_triangular(L, b, lower=self._lower)

  def tree_flatten(self) -> Tuple[Sequence[Any], Dict[str, Any]]:
    children, aux_data = super().tree_flatten()
    aux_data["lower"] = self._lower
    return children, aux_data


@jax.tree_util.register_pytree_node_class
class SparseCholeskySolver(
    CholeskySolver[Union[jesp.CSR, jesp.CSC, jesp.COO, jesp.BCOO]]
):
  r"""Sparse Cholesky solver using :func:`jax.experimental.host_callback.call`.

  Uses the CHOLMOD :cite:`cholmod:08` bindings from :mod:`sksparse.cholmod`.

  Args:
    A: Symmetric positive definite matrix of shape ``[n, n]``.
    beta: Decompose :math:`A + \beta * I` instead of :math:`A`.
    key: Key used to cache :class:`sksparse.cholmod.Factor`.
      This key **must** be unique to ``A`` to achieve correct results.
      If `None`, use :func:`hash` of this object.
    kwargs: Keyword arguments for :func:`sksparse.cholmod.cholesky`.
  """

  # TODO(michalk8): in the future, define a jax primitive + use CHOLMOD directly
  _FACTOR_CACHE = {}

  def __init__(
      self,
      A: T,
      beta: float = 0.0,
      key: Optional[Hashable] = None,
      **kwargs: Any,
  ):
    if cholmod is None:
      raise ImportError(
          "Unable to import scikit-sparse. "
          "Please install it as `pip install scikit-sparse`."
      )
    super().__init__(A)
    self._key = key
    self._beta = beta
    self._kwargs = kwargs

  def _host_decompose(self, A: T) -> None:
    # use float64 because CHOLMOD uses it internally
    # convert to CSC explicitly for efficiency/to avoid warnings
    mat = _jax_sparse_to_scipy(A, sum_duplicates=True, dtype=float).tocsc()
    self._FACTOR_CACHE[hash(self)] = cholmod.cholesky(
        mat, beta=self._beta, **self._kwargs
    )

  def _decompose(self, A: T) -> Optional[T]:
    return hcb.call(self._host_decompose, A, result_shape=None)

  def _host_solve(self, b: jnp.ndarray) -> jnp.ndarray:
    factor = self._FACTOR_CACHE[hash(self)]
    return factor.solve_A(np.asarray(b, dtype=float))

  def _solve(self, _: Optional[T], b: jnp.ndarray) -> jnp.ndarray:
    return hcb.call(self._host_solve, b, result_shape=b)

  @property
  def L(self) -> None:
    """Compute the lower-triangular factor of :attr:`A` and cache the result.

    The factor is not returned, but is used in subsequent :meth:`solve` calls.
    """
    if hash(self) not in self._FACTOR_CACHE:
      self._decompose(jax.lax.stop_gradient(self.A))

  @classmethod
  def clear_factor_cache(cls) -> None:
    """Clear the :class:`sksparse.cholmod.Factor` cache."""
    cls._FACTOR_CACHE.clear()

  def __hash__(self) -> int:
    return object.__hash__(self) if self._key is None else self._key

  def tree_flatten(self) -> Tuple[Sequence[Any], Dict[str, Any]]:
    children, aux_data = super().tree_flatten()
    return children, {
        **aux_data, "beta": self._beta,
        "key": self._key,
        **self._kwargs
    }


def _jax_sparse_to_scipy(
    A: Union[jesp.CSR, jesp.CSC, jesp.COO],
    sum_duplicates: bool = False,
    **kwargs: Any
) -> sp.spmatrix:
  toarr = np.asarray

  if isinstance(A, (jesp.CSR, jesp.CSC)):
    data, indices, indptr = toarr(A.data), toarr(A.indices), toarr(A.indptr)
    return sp.csr_matrix((data, indices, indptr), **kwargs)
  if isinstance(A, jesp.COO):
    row, col, data = toarr(A.row), toarr(A.col), toarr(A.data)
    return sp.coo_matrix((data, (row, col)), **kwargs)
  if isinstance(A, jesp.BCOO):
    assert A.indices.ndim == 2, "Only 2D batched COO matrix is supported."
    row, col = A.indices[:, 0], A.indices[:, 1]
    data, row, col = toarr(A.data), toarr(row), toarr(col)
    mat = sp.coo_matrix((data, (row, col)), **kwargs)
    if not sum_duplicates:
      return mat

    # the original matrix can contain duplicates => shape will be wrong
    # we optionally correct it here
    mat.sum_duplicates()
    mat.eliminate_zeros()

    return sp.coo_matrix((mat.data, (mat.row, mat.col)), **kwargs)

  raise TypeError(type(A))


def _scipy_sparse_to_jax(A: sp.spmatrix,
                         **kwargs: Any) -> Union[jesp.CSR, jesp.CSC, jesp.BCOO]:
  toarr = jnp.asarray
  kwargs["shape"] = A.shape

  if sp.isspmatrix_csr(A):
    return jesp.CSR((toarr(A.data), toarr(A.indices), toarr(A.indptr)),
                    **kwargs)
  if sp.isspmatrix_csc(A):
    return jesp.CSC((toarr(A.data), toarr(A.indices), toarr(A.indptr)),
                    **kwargs)
  if sp.isspmatrix_coo(A):
    # prefer BCOO since it's more feature-complete
    data, indices = toarr(A.data), jnp.c_[toarr(A.row), toarr(A.col)]
    return jesp.BCOO((data, indices), **kwargs)

  raise TypeError(type(A))



# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""A class describing low-rank geometries."""
from typing import Any, Callable, Optional, Tuple, Union

import jax
import jax.numpy as jnp
from typing_extensions import Literal

from ott.geometry import geometry

__all__ = ["LRCGeometry"]


@jax.tree_util.register_pytree_node_class
class LRCGeometry(geometry.Geometry):
  """Geometry whose cost is defined by product of two low-rank matrices.

  Implements geometries that are defined as low rank products, i.e. for which
  there exists two matrices :math:`A` and :math:`B` of :math:`r` columns such
  that the cost of the geometry equals :math:`AB^T`. Apart from being faster to
  apply to a vector, these geometries are characterized by the fact that adding
  two such geometries should be carried out by concatenating factors, i.e.
  if :math:`C = AB^T` and :math:`D = EF^T` then :math:`C + D = [A,E][B,F]^T`

  Args:
    cost_1: jnp.ndarray<float>[num_a, r]
    cost_2: jnp.ndarray<float>[num_b, r]
    bias: constant added to entire cost matrix.
    scale: Value used to rescale the factors of the low-rank geometry.
    scale_cost: option to rescale the cost matrix. Implemented scalings are
      'max_bound', 'mean' and 'max_cost'. Alternatively, a float
      factor can be given to rescale the cost such that
      ``cost_matrix /= scale_cost``. If `True`, use 'mean'.
    batch_size: optional size of the batch to compute online (without
      instantiating the matrix) the scale factor ``scale_cost`` of the
      :attr:`cost_matrix` when ``scale_cost = 'max_cost'``. If `None`, the batch
      size is set to `1024` or to the largest number of samples between
      :attr:`cost_1` and :attr:`cost_2` if smaller than `1024`.
    kwargs: Additional keyword arguments for
      :class:`~ott.geometry.geometry.Geometry`.
  """

  def __init__(
      self,
      cost_1: jnp.ndarray,
      cost_2: jnp.ndarray,
      bias: float = 0.,
      scale_factor: float = 1.,
      scale_cost: Union[bool, int, float, Literal['mean', 'max_bound',
                                                  'max_cost']] = 1.0,
      batch_size: Optional[int] = None,
      **kwargs: Any,
  ):
    assert cost_1.shape[1] == cost_2.shape[1]
    self._cost_1 = cost_1
    self._cost_2 = cost_2
    self._bias = bias
    self._scale_factor = scale_factor
    self._kwargs = kwargs

    super().__init__(**kwargs)
    self._scale_cost = 'mean' if scale_cost is True else scale_cost
    self.batch_size = batch_size

  @property
  def cost_1(self) -> jnp.ndarray:
    """First factor of the :attr:`cost_matrix`."""
    scale_factor = jnp.sqrt(self._scale_factor * self.inv_scale_cost)
    return scale_factor * self._cost_1

  @property
  def cost_2(self) -> jnp.ndarray:
    """Second factor of the :attr:`cost_matrix`."""
    scale_factor = jnp.sqrt(self._scale_factor * self.inv_scale_cost)
    return scale_factor * self._cost_2

  @property
  def bias(self) -> float:
    """Constant offset added to the entire :attr:`cost_matrix`."""
    return self._bias * self.inv_scale_cost

  @property
  def cost_rank(self) -> int:
    return self._cost_1.shape[1]

  @property
  def cost_matrix(self) -> jnp.ndarray:
    """Materialize the cost matrix."""
    return jnp.matmul(self.cost_1, self.cost_2.T) + self.bias

  @property
  def shape(self) -> Tuple[int, int]:
    return self._cost_1.shape[0], self._cost_2.shape[0]

  @property
  def is_symmetric(self) -> bool:
    return (
        self._cost_1.shape[0] == self._cost_2.shape[0] and
        jnp.all(self._cost_1 == self._cost_2)
    )

  @property
  def inv_scale_cost(self) -> float:
    if isinstance(self._scale_cost, (int, float, jnp.DeviceArray)):
      return 1.0 / self._scale_cost
    self = self._masked_geom()
    if self._scale_cost == 'max_bound':
      x_norm = self._cost_1[:, 0].max()
      y_norm = self._cost_2[:, 1].max()
      max_bound = x_norm + y_norm + 2 * jnp.sqrt(x_norm * y_norm)
      return 1.0 / (max_bound + self._bias)
    if self._scale_cost == 'mean':
      factor1 = jnp.dot(self._n_normed_ones, self._cost_1)
      factor2 = jnp.dot(self._cost_2.T, self._m_normed_ones)
      mean = jnp.dot(factor1, factor2) + self._bias
      return 1.0 / mean
    if self._scale_cost == 'max_cost':
      return 1.0 / self.compute_max_cost()
    raise ValueError(f'Scaling {self._scale_cost} not implemented.')

  def apply_square_cost(self, arr: jnp.ndarray, axis: int = 0) -> jnp.ndarray:
    """Apply elementwise-square of cost matrix to array (vector or matrix)."""
    (n, m), r = self.shape, self.cost_rank
    # When applying square of a LRCGeometry, one can either elementwise square
    # the cost matrix, or instantiate an augmented (rank^2) LRCGeometry
    # and apply it. First is O(nm), the other is O((n+m)r^2).
    if n * m < (n + m) * r ** 2:  # better use regular apply
      return super().apply_square_cost(arr, axis)
    else:
      new_cost_1 = self.cost_1[:, :, None] * self.cost_1[:, None, :]
      new_cost_2 = self.cost_2[:, :, None] * self.cost_2[:, None, :]
      return LRCGeometry(
          cost_1=new_cost_1.reshape((n, r ** 2)),
          cost_2=new_cost_2.reshape((m, r ** 2))
      ).apply_cost(arr, axis)

  def _apply_cost_to_vec(
      self,
      vec: jnp.ndarray,
      axis: int = 0,
      fn: Optional[Callable[[jnp.ndarray], jnp.ndarray]] = None,
      is_linear: bool = False,
  ) -> jnp.ndarray:
    """Apply [num_a, num_b] fn(cost) (or transpose) to vector.

    Args:
      vec: jnp.ndarray [num_a,] ([num_b,] if axis=1) vector
      axis: axis on which the reduction is done.
      fn: function optionally applied to cost matrix element-wise, before the
        doc product
      is_linear: Whether ``fn`` is a linear function to enable efficient
        implementation. See :func:`ott.geometry.geometry.is_linear`
        for a heuristic to help determine if a function is linear.

    Returns:
      A jnp.ndarray corresponding to cost x vector
    """

    def linear_apply(
        vec: jnp.ndarray, axis: int, fn: Callable[[jnp.ndarray], jnp.ndarray]
    ) -> jnp.ndarray:
      c1 = self.cost_1 if axis == 1 else self.cost_2
      c2 = self.cost_2 if axis == 1 else self.cost_1
      c2 = fn(c2) if fn is not None else c2
      bias = fn(self.bias) if fn is not None else self.bias
      out = jnp.dot(c1, jnp.dot(c2.T, vec))
      return out + bias * jnp.sum(vec) * jnp.ones_like(out)

    if fn is None or is_linear:
      return linear_apply(vec, axis, fn=fn)
    return super()._apply_cost_to_vec(vec, axis, fn=fn)

  def compute_max_cost(self) -> float:
    """Compute the maximum of the :attr:`cost_matrix`.

    Three cases are taken into account:

    - If the number of samples of ``cost_1`` and ``cost_2`` are both smaller
      than 1024 and if ``batch_size`` is `None`, the ``cost_matrix`` is
      computed to obtain its maximum entry.
    - If one of the number of samples of ``cost_1`` or ``cost_2`` is larger
      than 1024 and if ``batch_size`` is `None`, then the maximum of the
      cost matrix is calculated by batch. The batches are created on the
      longest axis of the cost matrix and their size is fixed to 1024.
    - If ``batch_size`` is provided as a float, then the maximum of the cost
      matrix is calculated by batch. The batches are created on the longest
      axis of the cost matrix and their size if fixed by ``batch_size``.

    Returns:
      Maximum of the cost matrix.
    """
    batch_for_y = self.shape[1] > self.shape[0]

    n = self.shape[1] if batch_for_y else self.shape[0]
    p = self._cost_2.shape[1] if batch_for_y else self._cost_1.shape[1]
    carry = ((self._cost_1, self._cost_2) if batch_for_y else
             (self._cost_2, self._cost_1))

    if self.batch_size:
      batch_size = min(self.batch_size, n)
    else:
      batch_size = min(1024, max(self.shape[0], self.shape[1]))
    n_batch = n // batch_size

    def body(carry, slice_idx):
      cost1, cost2 = carry
      cost2_slice = jax.lax.dynamic_slice(
          cost2, (slice_idx * batch_size, 0), (batch_size, p)
      )
      out_slice = jnp.max(jnp.dot(cost2_slice, cost1.T))
      return carry, out_slice

    def finalize(carry):
      cost1, cost2 = carry
      out_slice = jnp.dot(cost2[n_batch * batch_size:], cost1.T)
      return out_slice

    _, out = jax.lax.scan(body, carry, jnp.arange(n_batch))
    last_slice = finalize(carry)
    max_value = jnp.max(jnp.concatenate((out, last_slice.reshape(-1))))
    return max_value + self._bias

  def to_LRCGeometry(
      self, rank: int, tol: float = 1e-2, seed: int = 0
  ) -> 'LRCGeometry':
    """Return self."""
    return self

  def subset(
      self, src_ixs: Optional[jnp.ndarray], tgt_ixs: Optional[jnp.ndarray],
      **kwargs: Any
  ) -> "LRCGeometry":

    def subset_fn(
        arr: Optional[jnp.ndarray],
        ixs: Optional[jnp.ndarray],
    ) -> jnp.ndarray:
      return arr if arr is None or ixs is None else arr[jnp.atleast_1d(ixs)]

    return self._mask_subset_helper(
        src_ixs, tgt_ixs, fn=subset_fn, propagate_mask=True, **kwargs
    )

  def mask(
      self,
      src_mask: Optional[jnp.ndarray],
      tgt_mask: Optional[jnp.ndarray],
      mask_value: float = 0.,
  ) -> "LRCGeometry":

    def mask_fn(
        arr: Optional[jnp.ndarray],
        mask: Optional[jnp.ndarray],
    ) -> Optional[jnp.ndarray]:
      if arr is None or mask is None:
        return arr
      return jnp.where(mask[:, None], arr, mask_value)

    src_mask = self._normalize_mask(src_mask, self.shape[0])
    tgt_mask = self._normalize_mask(tgt_mask, self.shape[1])
    return self._mask_subset_helper(
        src_mask, tgt_mask, fn=mask_fn, propagate_mask=False
    )

  def _mask_subset_helper(
      self,
      src_ixs: Optional[jnp.ndarray],
      tgt_ixs: Optional[jnp.ndarray],
      *,
      fn: Callable[[Optional[jnp.ndarray], Optional[jnp.ndarray]],
                   Optional[jnp.ndarray]],
      propagate_mask: bool,
      **kwargs: Any,
  ) -> "LRCGeometry":
    (c1, c2, src_mask, tgt_mask, *children), aux_data = self.tree_flatten()
    c1 = fn(c1, src_ixs)
    c2 = fn(c2, tgt_ixs)
    if propagate_mask:
      src_mask = self._normalize_mask(src_mask, self.shape[0])
      tgt_mask = self._normalize_mask(tgt_mask, self.shape[1])
      src_mask = fn(src_mask, src_ixs)
      tgt_mask = fn(tgt_mask, tgt_ixs)

    aux_data = {**aux_data, **kwargs}
    return type(self).tree_unflatten(
        aux_data, [c1, c2, src_mask, tgt_mask] + children
    )

  def __add__(self, other: 'LRCGeometry') -> 'LRCGeometry':
    assert isinstance(other, LRCGeometry), type(other)
    return type(self)(
        cost_1=jnp.concatenate((self.cost_1, other.cost_1), axis=1),
        cost_2=jnp.concatenate((self.cost_2, other.cost_2), axis=1),
        **self._kwargs
    )

  @property
  def dtype(self) -> jnp.dtype:
    return self._cost_1.dtype

  def tree_flatten(self):
    return (
        self._cost_1, self._cost_2, self._src_mask, self._tgt_mask, self._kwargs
    ), {
        'bias': self._bias,
        'scale_factor': self._scale_factor,
        'scale_cost': self._scale_cost,
        'batch_size': self.batch_size
    }

  @classmethod
  def tree_unflatten(cls, aux_data, children):
    c1, c2, src_mask, tgt_mask, kwargs = children
    return cls(
        c1, c2, src_mask=src_mask, tgt_mask=tgt_mask, **kwargs, **aux_data
    )



from typing import Callable, NamedTuple

import jax
import jax.numpy as jnp

__all__ = ["make_square_loss", "make_kl_loss"]


class Loss(NamedTuple):
  func: Callable[[jnp.ndarray], jnp.ndarray]
  is_linear: bool


class GWLoss(NamedTuple):
  f1: Loss
  f2: Loss
  h1: Loss
  h2: Loss


def make_square_loss() -> GWLoss:
  f1 = Loss(lambda x: x ** 2, is_linear=False)
  f2 = Loss(lambda y: y ** 2, is_linear=False)
  h1 = Loss(lambda x: x, is_linear=True)
  h2 = Loss(lambda y: 2.0 * y, is_linear=True)
  return GWLoss(f1, f2, h1, h2)


def make_kl_loss(clipping_value: float = 1e-8) -> GWLoss:
  f1 = Loss(lambda x: -jax.scipy.special.entr(x) - x, is_linear=False)
  f2 = Loss(lambda y: y, is_linear=True)
  h1 = Loss(lambda x: x, is_linear=True)
  h2 = Loss(lambda y: jnp.log(jnp.clip(y, clipping_value)), is_linear=False)
  return GWLoss(f1, f2, h1, h2)



# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Pytree containing parameters for a pair of coupled Gaussian mixture models.
"""  # noqa: D200
import jax
import jax.numpy as jnp

from ott.geometry import costs, geometry, pointcloud
from ott.solvers.linear import sinkhorn
from ott.tools.gaussian_mixture import gaussian_mixture


@jax.tree_util.register_pytree_node_class
class GaussianMixturePair:
  """Pytree for a coupled pair of Gaussian mixture models.

  Includes methods used in estimating an optimal pairing between GMM components
  using the Wasserstein-like method described in :cite:`delon:20`,
  as well as generalization that allows for the re-weighting of components.

  The Delon & Desolneux paper above proposes fitting a pair of GMMs to a pair
  of point clouds in such a way that the sum of the log likelihood of the
  points minus a weighted penalty involving a Wasserstein-like distance between
  the GMMs. Their proposed algorithm involves using EM in which a balanced
  Sinkhorn algorithm is used to estimate a coupling between the GMMs at each
  step of EM.

  Our generalization of this algorithm allows for a mismatch between the
  marginals of the coupling and the GMM component weights. This mismatch can be
  interpreted as components being re-weighted rather than being transported.
  We penalize reweighting with a generalized KL-divergence penalty, and we give
  the option to use the unbalanced Sinkhorn algorithm rather than the balanced
  to compute the divergence between GMMs.
  """

  def __init__(
      self,
      gmm0: gaussian_mixture.GaussianMixture,
      gmm1: gaussian_mixture.GaussianMixture,
      epsilon: float = 1.e-2,
      tau: float = 1.,
      lock_gmm1: bool = False,
  ):
    """Constructor.

    When fitting a pair of coupled GMMs with *no* reweighting of components
    using the algorithm in :cite:`delon:20`, set tau = 1. The coupling between
    components will be determined via the balanced Sinkhorn algorithm.

    When fitting a pair of coupled GMMs in which reweighting of components is
    allowed, set tau to a value in (0, 1). The resulting coupling will penalize
    the generalized KL divergence between the coupling's marginals and the GMM
    component weights with a weight of rho = epsilon tau / (1 - tau).

    Args:
      gmm0: first GMM in the pair
      gmm1: second GMM in the pair
      epsilon: regularization weight to use for the Sinkhorn algorithm
      tau: encodes the weight, rho, to use for the generalized KL divergence
        between the coupling's marginals and GMM component weights as
        rho = epsilon tau / (1 - tau)
      lock_gmm1: indicates whether the parameters of gmm1 should be modified
        during optimization
    """  # noqa: D401
    self._gmm0 = gmm0
    self._gmm1 = gmm1
    self._epsilon = epsilon
    self._tau = tau
    self._lock_gmm1 = lock_gmm1

  @property
  def dtype(self):
    return self.gmm0.dtype

  @property
  def gmm0(self):
    return self._gmm0

  @property
  def gmm1(self):
    return self._gmm1

  @property
  def epsilon(self):
    return self._epsilon

  @property
  def tau(self):
    return self._tau

  @property
  def rho(self):
    return self.epsilon * self.tau / (1. - self.tau)

  @property
  def lock_gmm1(self):
    return self._lock_gmm1

  def get_bures_geometry(self) -> pointcloud.PointCloud:
    """Get a Bures Geometry for the two GMMs."""
    mean0 = self.gmm0.loc
    dimension = mean0.shape[-1]
    cov0 = self.gmm0.covariance
    cov0 = cov0.reshape(cov0.shape[:-2] + (dimension * dimension,))
    x = jnp.concatenate([mean0, cov0], axis=-1)
    mean1 = self.gmm1.loc
    cov1 = self.gmm1.covariance
    cov1 = cov1.reshape(cov1.shape[:-2] + (dimension * dimension,))
    y = jnp.concatenate([mean1, cov1], axis=-1)
    return pointcloud.PointCloud(
        x=x,
        y=y,
        cost_fn=costs.Bures(dimension=dimension),
        epsilon=self.epsilon
    )

  def get_cost_matrix(self) -> jnp.ndarray:
    """Get matrix of W2^2 costs between all pairs of (gmm0, gmm1) components."""
    return self.get_bures_geometry().cost_matrix

  def get_sinkhorn(self, cost_matrix: jnp.ndarray) -> sinkhorn.SinkhornOutput:
    """Get the output of ott.sinkhorn's method for a given cost matrix."""
    # We use a Geometry here rather than the PointCloud created in
    # in get_bures_geometry to avoid recomputing the cost matrix, since
    # the cost matrix is quite expensive
    geom = geometry.Geometry(cost_matrix=cost_matrix, epsilon=self.epsilon)
    return sinkhorn.sinkhorn(
        geom,
        a=self.gmm0.component_weights,
        b=self.gmm1.component_weights,
        tau_a=self.tau,
        tau_b=self.tau
    )

  def get_normalized_sinkhorn_coupling(
      self,
      sinkhorn_output: sinkhorn.SinkhornOutput,
  ) -> jnp.ndarray:
    """Get the normalized coupling matrix for the specified Sinkhorn output.

    Args:
      sinkhorn_output: Sinkhorn algorithm output as returned by get_sinkhorn()

    Returns:
      A coupling matrix that tells how much of the mass of each component of
      gmm0 is mapped to each component of gmm1.
    """
    return sinkhorn_output.matrix / jnp.sum(sinkhorn_output.matrix)

  def tree_flatten(self):
    """Method used by jax.tree_util to flatten a GaussianMixturePair.

    We control the subset of parameters that we will optimize in fit_gmm_pair
    by selectively placing them in either children (the parameters to optimize)
    or aux_data (the parameters to leave alone).

    Returns:
      A tuple of child pytrees and a dict of auxiliary data.
    """  # noqa: D401
    children = [self.gmm0]
    aux_data = {
        'epsilon': self.epsilon,
        'tau': self.tau,
        'lock_gmm1': self.lock_gmm1
    }
    if self.lock_gmm1:
      aux_data['gmm1'] = self.gmm1
    else:
      children.append(self.gmm1)
    return tuple(children), aux_data

  @classmethod
  def tree_unflatten(cls, aux_data, children):
    """Method used by jax.tree_util to unflatten a GaussianMixturePair.

    tree_flatten controls which parameters get optimized by placing them in
    either children or aux_data; here we invert the process.

    Args:
      aux_data: auxiliary data that is passed to the constructor as kwargs
      children: child pytrees passed to the constructor as args

    Returns:
      A GaussianMixturePair.
    """  # noqa: D401
    children = list(children)
    if 'gmm1' in aux_data:
      gmm1 = aux_data.pop('gmm1')
      children.insert(1, gmm1)
    return cls(*children, **aux_data)

  def __repr__(self):
    class_name = type(self).__name__
    children, aux = self.tree_flatten()
    return '{}({})'.format(
        class_name, ', '.join([repr(c) for c in children] +
                              [f'{k}: {repr(v)}' for k, v in aux.items()])
    )

  def __hash__(self):
    return jax.tree_util.tree_flatten(self).__hash__()

  def __eq__(self, other):
    return jax.tree_util.tree_flatten(self) == jax.tree_util.tree_flatten(other)



# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Fit a Gaussian mixture model.

Sample usage:

# initialize GMM with K-means++
gmm_init = fit_gmm.initialize(
  key=key,
  points=my_points,
  point_weights=None,
  n_components=COMPONENTS)

# refine GMM parameters using EM
gmm = fit_gmm.fit_model_em(
  gmm=gmm_init,
  points=my_points,
  point_weights=None,
  steps=10,
  verbose=True)


We fit the model using EM. Below we'll use notation following
https://en.wikipedia.org/wiki/Expectation%E2%80%93maximization_algorithm

Our data X is generated by a Gaussian mixture with unknown parameters \Theta.
We denote the (unobserved) component that gave rise to each point as Z.

In EM we start with an initial estimate of $\Theta$, $\Theta^{(0)}$, and we
then iteratively update it via

$$\Theta^{(t+1)} = \argmax_{\Theta} Q(\Theta|\Theta^{(t)})$$

where

$$
Q(\Theta|\Theta(t)) = E_{Z|X,\Theta^{(t)}} \left[ \log L(\Theta; X, Z) \right]
$$
"""

from typing import Optional

import jax
import jax.numpy as jnp

from ott.tools.gaussian_mixture import gaussian_mixture

# EM algorithm for parameter estimation


def get_assignment_probs(
    gmm: gaussian_mixture.GaussianMixture, points: jnp.ndarray
) -> jnp.ndarray:
  r"""Get component assignment probabilities used in the E step of EM.

  Here we compute the component assignment probabilities p(Z|X, \Theta^{(t)})
  that we need to compute the expectation used for Q(\Theta|\Theta^{(t)}).

  Args:
    gmm: GMM model
    points: set of samples being fitted, shape (n, n_dimensions)

  Returns:
    An array of assignment probabilities with shape (n, n_components)
  """
  return jnp.exp(gmm.get_log_component_posterior(points))


def get_q(
    gmm: gaussian_mixture.GaussianMixture,
    assignment_probs: jnp.ndarray,
    points: jnp.ndarray,
    point_weights: Optional[jnp.ndarray] = None,
) -> float:
  r"""Get Q(\Theta|\Theta^{(t)}).

  Args:
    gmm: GaussianMixture with parameters \Theta
    assignment_probs: p(Z|X, \Theta^{(t)}) as computed by get_assignment_probs
    points: observations X
    point_weights: optional set of weights for the samples. If None, use
      a weight of 1/n where n is the number of points.

  Returns:
    Q(\Theta|\Theta^{(t)})
  """
  # log P(X, Z| \Theta) = log P(X|Z, \Theta) + log P(Z|\Theta)
  loglik = (gmm.conditional_log_prob(points) + gmm.log_component_weights())
  if point_weights is None:
    point_weights = jnp.ones(points.shape[0])
  return (
      jnp.sum(point_weights * jnp.sum(assignment_probs * loglik, axis=-1)) /
      jnp.sum(point_weights)
  )


def log_prob_loss(
    gmm: gaussian_mixture.GaussianMixture,
    points: jnp.ndarray,
    point_weights: Optional[jnp.ndarray] = None,
) -> float:
  """Loss function: weighted mean of (-log prob of observations).

  Args:
    gmm: GMM model
    points: set of samples being fitted
    point_weights: optional set of weights for the samples. If None, use
      a weight of 1/n where n is the number of points.

  Returns:
    The GMM loss for the points.
  """
  if point_weights is None:
    return -jnp.mean(gmm.log_prob(points))
  return -jnp.sum(point_weights * gmm.log_prob(points)) / jnp.sum(point_weights)


def fit_model_em(
    gmm: gaussian_mixture.GaussianMixture,
    points: jnp.ndarray,
    point_weights: Optional[jnp.ndarray],
    steps: int,
    jit: bool = True,
    verbose: bool = False,
) -> gaussian_mixture.GaussianMixture:
  """Fit a GMM using the EM algorithm.

  Args:
    gmm: initial GMM model
    points: set of samples to fit, shape (n, n_dimensions)
    point_weights: optional set of weights for points, shape (n,). If None,
      uses equal weights for all points.
    steps: number of steps of EM to perform
    jit: if True, compile functions
    verbose: if True, print the loss at each step

  Returns:
    A GMM with updated parameters.
  """
  if point_weights is None:
    point_weights = jnp.ones(points.shape[:-1], dtype=points.dtype)
  loss_fn = log_prob_loss
  get_q_fn = get_q
  e_step_fn = get_assignment_probs
  m_step_fn = gaussian_mixture.GaussianMixture.from_points_and_assignment_probs
  if jit:
    loss_fn = jax.jit(loss_fn)
    get_q_fn = jax.jit(get_q_fn)
    e_step_fn = jax.jit(e_step_fn)
    m_step_fn = jax.jit(m_step_fn)

  for i in range(steps):
    assignment_probs = e_step_fn(gmm, points)
    gmm_new = m_step_fn(points, point_weights, assignment_probs)
    if gmm_new.has_nans():
      raise ValueError('NaNs in fit.')
    if verbose:
      loss = loss_fn(gmm_new, points, point_weights)
      q = get_q_fn(
          gmm=gmm_new,
          assignment_probs=assignment_probs,
          points=points,
          point_weights=point_weights
      )
      print(f'{i}  q={q}  -log prob={loss}', flush=True)
    gmm = gmm_new
  return gmm


# KMeans++ for initialization
# See https://en.wikipedia.org/wiki/K-means%2B%2B for details


def _get_dist_sq(points: jnp.ndarray, loc: jnp.ndarray) -> jnp.ndarray:
  """Get the squared distance from each point to each loc."""

  def _dist_sq_one_loc(points: jnp.ndarray, loc: jnp.ndarray) -> jnp.ndarray:
    return jnp.sum((points - loc[None]) ** 2., axis=-1)

  dist_sq_fn = jax.vmap(_dist_sq_one_loc, in_axes=(None, 0), out_axes=1)
  return dist_sq_fn(points, loc)


def _get_locs(
    key: jnp.ndarray, points: jnp.ndarray, n_components: int
) -> jnp.ndarray:
  """Get the initial component means.

  Args:
    key: jax.random seed
    points: (n, n_dimensions) array of observations
    n_components: desired number of components

  Returns:
    (n_components, n_dimensions) array of means.
  """
  points = points.copy()
  n_points = points.shape[0]
  weights = jnp.ones(n_points) / n_points
  key, subkey = jax.random.split(key)
  index = jax.random.choice(key=subkey, a=points.shape[0], p=weights)
  loc = points[index]
  points = jnp.concatenate([points[:index], points[index + 1:]], axis=0)

  locs = loc[None]
  for _ in range(n_components - 1):
    dist_sq = _get_dist_sq(points, locs)
    min_dist_sq = jnp.min(dist_sq, axis=-1)
    weights = min_dist_sq / jnp.sum(min_dist_sq)
    key, subkey = jax.random.split(key)
    index = jax.random.choice(key=subkey, a=points.shape[0], p=weights)
    loc = points[index]
    points = jnp.concatenate([points[:index], points[index + 1:]], axis=0)
    locs = jnp.concatenate([locs, loc[None]], axis=0)
  return locs


def from_kmeans_plusplus(
    key: jnp.ndarray,
    points: jnp.ndarray,
    point_weights: Optional[jnp.ndarray],
    n_components: int,
) -> gaussian_mixture.GaussianMixture:
  """Initialize a GMM via a single pass of K-means++.

  Args:
    key: jax.random seed
    points: (n, n_dimensions) array of observations
    point_weights: (n,) array of weights for points
    n_components: desired number of components

  Returns:
    An initial Gaussian mixture model.

  Raises:
    ValueError if any fitted parameters are non-finite.
  """
  key, subkey = jax.random.split(key)
  locs = _get_locs(key=subkey, points=points, n_components=n_components)
  dist_sq = _get_dist_sq(points, locs)
  assignment_prob = (dist_sq == jnp.min(dist_sq,
                                        axis=-1)[:, None]).astype(points.dtype)
  del dist_sq

  if point_weights is None:
    point_weights = jnp.ones_like(points[..., 0])
  return gaussian_mixture.GaussianMixture.from_points_and_assignment_probs(
      points=points,
      point_weights=point_weights,
      assignment_probs=assignment_prob
  )


def initialize(
    key: jnp.ndarray,
    points: jnp.ndarray,
    point_weights: Optional[jnp.ndarray],
    n_components: int,
    n_attempts: int = 50,
    verbose: bool = False
) -> gaussian_mixture.GaussianMixture:
  """Initialize a GMM via K-means++ with retries on failure.

  Args:
    key: jax.random seed
    points: (n, n_dimensions) array of observations
    point_weights: (n,) array of weights for points
    n_components: desired number of components
    n_attempts: number of attempts to initialize before failing
    verbose: if True, print status information

  Returns:
    An initial Gaussian mixture model.

  Raises:
    ValueError if initialization was unsuccessful after n_attempts attempts.
  """
  for attempt in range(n_attempts):
    key, subkey = jax.random.split(key)
    try:
      return from_kmeans_plusplus(
          key=subkey,
          points=points,
          point_weights=point_weights,
          n_components=n_components
      )
    except ValueError:
      if verbose:
        print(f'Failed to initialize, attempt {attempt}.', flush=True)
  raise ValueError('Failed to initialize.')



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""A Jax implementation of the ICNN based Kantorovich dual."""

import warnings
from typing import Dict, Iterable, List, Optional, Tuple, Union

import flax.linen as nn
import jax
import jax.numpy as jnp
import optax
from flax import core
from typing_extensions import Literal

from ott.geometry import costs
from ott.problems.linear import potentials
from ott.solvers.nn import icnn

__all__ = ["NeuralDualSolver"]

Train_t = Dict[Literal["train_logs", "valid_logs"], Dict[str, List[float]]]


class NeuralDualSolver:
  r"""Solver of the ICNN-based Kantorovich dual.

  Learn the optimal transport between two distributions, denoted source
  and target, respectively. This is achieved by parameterizing the two
  Kantorovich potentials, `g` and `f`, by two input convex neural networks.
  :math:`\nabla g` hereby transports source to target cells, and
  :math:`\nabla f` target to source cells.

  Original algorithm is described in :cite:`makkuva:20`.

  Args:
    input_dim: input dimensionality of data required for network init
    neural_f: network architecture for potential f
    neural_g: network architecture for potential g
    optimizer_f: optimizer function for potential f
    optimizer_g: optimizer function for potential g
    num_train_iters: number of total training iterations
    num_inner_iters: number of training iterations of g per iteration of f
    valid_freq: frequency with which model is validated
    log_freq: frequency with training and validation are logged
    logging: option to return logs
    seed: random seed for network initializations
    pos_weights: option to train networks with positive weights or regularizer
    beta: regularization parameter when not training with positive weights
  """

  def __init__(
      self,
      input_dim: int,
      neural_f: Optional[nn.Module] = None,
      neural_g: Optional[nn.Module] = None,
      optimizer_f: Optional[optax.OptState] = None,
      optimizer_g: Optional[optax.OptState] = None,
      num_train_iters: int = 100,
      num_inner_iters: int = 10,
      valid_freq: int = 100,
      log_freq: int = 100,
      logging: bool = False,
      seed: int = 0,
      pos_weights: bool = True,
      beta: float = 1.0,
  ):
    self.num_train_iters = num_train_iters
    self.num_inner_iters = num_inner_iters
    self.valid_freq = valid_freq
    self.log_freq = log_freq
    self.logging = logging
    self.pos_weights = pos_weights
    self.beta = beta

    # set random key
    rng = jax.random.PRNGKey(seed)

    # set default optimizers
    if optimizer_f is None:
      optimizer_f = optax.adam(learning_rate=0.0001, b1=0.5, b2=0.9, eps=1e-8)
    if optimizer_g is None:
      optimizer_g = optax.adam(learning_rate=0.0001, b1=0.5, b2=0.9, eps=1e-8)

    # set default neural architectures
    if neural_f is None:
      neural_f = icnn.ICNN(dim_hidden=[64, 64, 64, 64])
    if neural_g is None:
      neural_g = icnn.ICNN(dim_hidden=[64, 64, 64, 64])

    # set optimizer and networks
    self.setup(rng, neural_f, neural_g, input_dim, optimizer_f, optimizer_g)

  def setup(
      self, rng: jnp.ndarray, neural_f: icnn.ICNN, neural_g: icnn.ICNN,
      input_dim: int, optimizer_f: optax.OptState, optimizer_g: optax.OptState
  ) -> None:
    """Setup all components required to train the network."""
    # split random key
    rng, rng_f, rng_g = jax.random.split(rng, 3)

    # check setting of network architectures
    if (
        neural_f.pos_weights != self.pos_weights or
        (neural_g.pos_weights != self.pos_weights)
    ):
      warnings.warn(
          f"Setting of ICNN and the positive weights setting of the "
          f"`NeuralDualSolver` are not consistent. Proceeding with "
          f"the `NeuralDualSolver` setting, with positive weights "
          f"being {self.pos_weights}."
      )
      neural_f.pos_weights = self.pos_weights
      neural_g.pos_weights = self.pos_weights

    self.state_f = neural_f.create_train_state(rng_f, optimizer_f, input_dim)
    self.state_g = neural_g.create_train_state(rng_g, optimizer_g, input_dim)

    # define train and valid step functions
    self.train_step_f = self.get_step_fn(train=True, to_optimize="f")
    self.valid_step_f = self.get_step_fn(train=False, to_optimize="f")

    self.train_step_g = self.get_step_fn(train=True, to_optimize="g")
    self.valid_step_g = self.get_step_fn(train=False, to_optimize="g")

  def __call__(
      self,
      trainloader_source: Iterable[jnp.ndarray],
      trainloader_target: Iterable[jnp.ndarray],
      validloader_source: Iterable[jnp.ndarray],
      validloader_target: Iterable[jnp.ndarray],
  ) -> Union[potentials.DualPotentials, Tuple[potentials.DualPotentials,
                                              Train_t]]:
    logs = self.train_neuraldual(
        trainloader_source,
        trainloader_target,
        validloader_source,
        validloader_target,
    )
    res = self.to_dual_potentials()

    return (res, logs) if self.logging else res

  def train_neuraldual(
      self,
      trainloader_source: Iterable[jnp.ndarray],
      trainloader_target: Iterable[jnp.ndarray],
      validloader_source: Iterable[jnp.ndarray],
      validloader_target: Iterable[jnp.ndarray],
  ) -> Train_t:
    """Implementation of the training and validation script."""  # noqa: D401
    try:
      from tqdm.auto import tqdm
    except ImportError:
      tqdm = lambda _: _
    # define dict to contain source and target batch
    batch_g = {}
    batch_f = {}
    valid_batch = {}

    # set logging dictionaries
    train_logs = {"train_loss_f": [], "train_loss_g": [], "train_w_dist": []}
    valid_logs = {"valid_loss_f": [], "valid_loss_g": [], "valid_w_dist": []}

    for step in tqdm(range(self.num_train_iters)):
      # execute training steps
      for _ in range(self.num_inner_iters):
        # get train batch for potential g
        batch_g["source"] = jnp.array(next(trainloader_source))
        batch_g["target"] = jnp.array(next(trainloader_target))

        self.state_g, loss_g, _ = self.train_step_g(
            self.state_f, self.state_g, batch_g
        )

      # get train batch for potential f
      batch_f["source"] = jnp.array(next(trainloader_source))
      batch_f["target"] = jnp.array(next(trainloader_target))

      self.state_f, loss_f, w_dist = self.train_step_f(
          self.state_f, self.state_g, batch_f
      )
      if not self.pos_weights:
        self.state_f = self.state_f.replace(
            params=self._clip_weights_icnn(self.state_f.params)
        )

      # log to wandb
      if self.logging and step % self.log_freq == 0:
        train_logs["train_loss_f"].append(float(loss_f))
        train_logs["train_loss_g"].append(float(loss_g))
        train_logs["train_w_dist"].append(float(w_dist))

      # report the loss on an validuation dataset periodically
      if step != 0 and step % self.valid_freq == 0:
        # get batch
        valid_batch["source"] = jnp.array(next(validloader_source))
        valid_batch["target"] = jnp.array(next(validloader_target))

        valid_loss_f, _ = self.valid_step_f(
            self.state_f, self.state_g, valid_batch
        )
        valid_loss_g, valid_w_dist = self.valid_step_g(
            self.state_f, self.state_g, valid_batch
        )

        if self.logging:
          # log training progress
          valid_logs["valid_loss_f"].append(float(valid_loss_f))
          valid_logs["valid_loss_g"].append(float(valid_loss_g))
          valid_logs["valid_w_dist"].append(float(valid_w_dist))

    return {"train_logs": train_logs, "valid_logs": valid_logs}

  def get_step_fn(self, train: bool, to_optimize: Literal["f", "g"] = "g"):
    """Create a one-step training and evaluation function."""

    def loss_fn(params_f, params_g, f, g, batch):
      """Loss function for potential f."""
      # get two distributions
      source, target = batch["source"], batch["target"]

      # get loss terms of kantorovich dual
      f_t = f({"params": params_f}, batch["target"])

      grad_g_s = jax.vmap(
          lambda x: jax.grad(g, argnums=1)({
              "params": params_g
          }, x)
      )(
          batch["source"]
      )

      f_grad_g_s = f({"params": params_f}, grad_g_s)

      s_dot_grad_g_s = jnp.sum(source * grad_g_s, axis=1)

      s_sq = jnp.sum(source * source, axis=1)
      t_sq = jnp.sum(target * target, axis=1)

      # compute final wasserstein distance
      dist = 2 * jnp.mean(
          f_grad_g_s - f_t - s_dot_grad_g_s + 0.5 * t_sq + 0.5 * s_sq
      )

      loss_f = jnp.mean(f_t - f_grad_g_s)
      loss_g = jnp.mean(f_grad_g_s - s_dot_grad_g_s)

      if to_optimize == "f":
        return loss_f, dist
      elif to_optimize == "g":
        if not self.pos_weights:
          penalty = self._penalize_weights_icnn(params_g)
          loss_g += self.beta * penalty
        return loss_g, dist
      else:
        raise ValueError("Optimization target has been misspecified.")

    @jax.jit
    def step_fn(state_f, state_g, batch):
      """Step function of either training or validation."""
      if to_optimize == "f":
        grad_fn = jax.value_and_grad(loss_fn, argnums=0, has_aux=True)
        state = state_f
      elif to_optimize == "g":
        grad_fn = jax.value_and_grad(loss_fn, argnums=1, has_aux=True)
        state = state_g
      else:
        raise ValueError("Potential to be optimize might be misspecified.")

      if train:
        # compute loss and gradients
        (loss, dist), grads = grad_fn(
            state_f.params,
            state_g.params,
            state_f.apply_fn,
            state_g.apply_fn,
            batch,
        )

        # update state
        return state.apply_gradients(grads=grads), loss, dist

      else:
        # compute loss and gradients
        (loss, dist), _ = grad_fn(
            state_f.params,
            state_g.params,
            state_f.apply_fn,
            state_g.apply_fn,
            batch,
        )

        # do not update state
        return loss, dist

    return step_fn

  def to_dual_potentials(self) -> potentials.DualPotentials:
    """Return the Kantorovich dual potentials from the trained potentials."""
    f = lambda x: self.state_f.apply_fn({"params": self.state_f.params}, x)
    g = lambda x: self.state_g.apply_fn({"params": self.state_g.params}, x)
    return potentials.DualPotentials(
        f, g, cost_fn=costs.SqEuclidean(), corr=True
    )

  @staticmethod
  def _clip_weights_icnn(params):
    params = params.unfreeze()
    for k in params.keys():
      if k.startswith("w_z"):
        params[k]["kernel"] = jnp.clip(params[k]["kernel"], a_min=0)

    return core.freeze(params)

  @staticmethod
  def _penalize_weights_icnn(
      params: Dict[str, jnp.ndarray]
  ) -> Dict[str, jnp.ndarray]:
    penalty = 0
    for k, param in params.items():
      if k.startswith("w_z"):
        penalty += jnp.linalg.norm(jax.nn.relu(-param["kernel"]))
    return penalty



import abc
from typing import TYPE_CHECKING, Any, Dict, Sequence, Tuple

import jax

from ott.geometry import geometry

if TYPE_CHECKING:
  from ott.initializers.linear import initializers_lr
  from ott.problems.linear import linear_problem
  from ott.problems.quadratic import quadratic_problem

__all__ = ["QuadraticInitializer", "LRQuadraticInitializer"]


@jax.tree_util.register_pytree_node_class
class BaseQuadraticInitializer(abc.ABC):
  """Base class for quadratic initializers.

  Args:
    kwargs: Keyword arguments.
  """

  def __init__(self, **kwargs: Any):
    self._kwargs = kwargs

  def __call__(
      self, quad_prob: 'quadratic_problem.QuadraticProblem', **kwargs: Any
  ) -> 'linear_problem.LinearProblem':
    """Compute the initial linearization of a quadratic problem.

    Args:
      quad_prob: Quadratic problem to linearize.
      kwargs: Additional keyword arguments.

    Returns:
      Linear problem.
    """
    from ott.problems.linear import linear_problem

    n, m = quad_prob.geom_xx.shape[0], quad_prob.geom_yy.shape[0]
    geom = self._create_geometry(quad_prob, **kwargs)
    assert geom.shape == (n, m), f"Expected geometry of shape `{n, m}`, " \
                                 f"found `{geom.shape}`."
    return linear_problem.LinearProblem(
        geom,
        a=quad_prob.a,
        b=quad_prob.b,
        tau_a=quad_prob.tau_a,
        tau_b=quad_prob.tau_b
    )

  @abc.abstractmethod
  def _create_geometry(
      self, quad_prob: 'quadratic_problem.QuadraticProblem', **kwargs: Any
  ) -> geometry.Geometry:
    """Compute initial geometry for linearization.

    Args:
      quad_problem: Quadratic problem.
      kwargs: Additional keyword arguments.

    Returns:
      Geometry used to initialize the linearized problem.
    """

  def tree_flatten(self) -> Tuple[Sequence[Any], Dict[str, Any]]:
    return [], self._kwargs

  @classmethod
  def tree_unflatten(
      cls, aux_data: Dict[str, Any], children: Sequence[Any]
  ) -> "BaseQuadraticInitializer":
    return cls(*children, **aux_data)


class QuadraticInitializer(BaseQuadraticInitializer):
  """Initialize a linear problem locally around a naive initializer ab'.

  If the problem is balanced (``tau_a = 1`` and ``tau_b = 1``),
  the equation of the cost follows eq. 6, p. 1 of :cite:`peyre:16`.

  If the problem is unbalanced (``tau_a < 1`` or ``tau_b < 1``), there are two
  possible cases. A first possibility is to introduce a quadratic KL
  divergence on the marginals in the objective as done in :cite:`sejourne:21`
  (``gw_unbalanced_correction = True``), which in turns modifies the
  local cost matrix.

  Alternatively, it could be possible to leave the formulation of the
  local cost unchanged, i.e. follow eq. 6, p. 1 of :cite:`peyre:16`
  (``gw_unbalanced_correction = False``) and include the unbalanced terms
  at the level of the linear problem only.

  Let :math:`P` [num_a, num_b] be the transport matrix, `cost_xx` is the
  cost matrix of `geom_xx` and `cost_yy` is the cost matrix of `geom_yy`.
  `left_x` and `right_y` depend on the loss chosen for GW.
  `gw_unbalanced_correction` is an boolean indicating whether or not the
  unbalanced correction applies.
  The equation of the local cost can be written as:

  `cost_matrix` = `marginal_dep_term`
              + `left_x`(`cost_xx`) :math:`P` `right_y`(`cost_yy`):math:`^T`
              + `unbalanced_correction` * `gw_unbalanced_correction`

  When working with the fused problem, a linear term is added to the cost
  matrix: `cost_matrix` += `fused_penalty` * `geom_xy.cost_matrix`
  """

  def _create_geometry(
      self, quad_prob: 'quadratic_problem.QuadraticProblem', *, epsilon: float,
      **kwargs: Any
  ) -> geometry.Geometry:
    """Compute initial geometry for linearization.

    Args:
      quad_prob: Quadratic OT problem.
      epsilon: Epsilon regularization.
      kwargs: Additional keyword arguments, unused.

    Returns:
      The initial geometry used to initialize the linearized problem.
    """
    from ott.problems.quadratic import quadratic_problem
    del kwargs

    unbalanced_correction = 0.0
    tmp = quad_prob.init_transport()
    marginal_1 = tmp.sum(1)
    marginal_2 = tmp.sum(0)

    # Initialises cost.
    marginal_cost = quad_prob.marginal_dependent_cost(marginal_1, marginal_2)

    if not quad_prob.is_balanced:
      transport_mass = marginal_1.sum()
      # Initialises epsilon for Unbalanced GW according to Sejourne et al (2021)
      epsilon = quadratic_problem.update_epsilon_unbalanced(
          epsilon=epsilon, transport_mass=transport_mass
      )
      unbalanced_correction = quad_prob.cost_unbalanced_correction(
          tmp, marginal_1, marginal_2, epsilon=epsilon
      )

    h1, h2 = quad_prob.quad_loss
    tmp = quadratic_problem.apply_cost(quad_prob.geom_xx, tmp, axis=1, fn=h1)
    tmp = quadratic_problem.apply_cost(
        quad_prob.geom_yy, tmp.T, axis=1, fn=h2
    ).T
    cost_matrix = (marginal_cost.cost_matrix - tmp + unbalanced_correction)

    cost_matrix += quad_prob.fused_penalty * quad_prob._fused_cost_matrix

    return geometry.Geometry(cost_matrix=cost_matrix, epsilon=epsilon)


class LRQuadraticInitializer(BaseQuadraticInitializer):
  """Wrapper that wraps low-rank Sinkhorn initializers.

  Args:
    lr_linear_initializer: Low-rank linear initializer.
  """

  def __init__(self, lr_linear_initializer: 'initializers_lr.LRInitializer'):
    super().__init__()
    self._linear_lr_initializer = lr_linear_initializer

  def _create_geometry(
      self, quad_prob: 'quadratic_problem.QuadraticProblem', **kwargs: Any
  ) -> geometry.Geometry:
    """Compute initial geometry for linearization.

    Args:
      quad_prob: Quadratic OT problem.
      kwargs: Keyword arguments for
        :meth:`~ott.initializers.linear.initializers_lr.LRInitializer.__call__`.

    Returns:
      The initial geometry used to initialize a linear problem.
    """
    from ott.solvers.linear import sinkhorn_lr

    q, r, g = self._linear_lr_initializer(quad_prob, **kwargs)
    tmp_out = sinkhorn_lr.LRSinkhornOutput(
        q=q, r=r, g=g, costs=None, errors=None, ot_prob=None
    )

    return quad_prob.update_lr_geom(tmp_out)

  @property
  def rank(self) -> int:
    """Rank of the transport matrix factorization."""
    return self._linear_lr_initializer.rank

  def tree_flatten(self) -> Tuple[Sequence[Any], Dict[str, Any]]:
    children, aux_data = super().tree_flatten()
    return children + [self._linear_lr_initializer], aux_data



# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Linear algebra utility methods for optimal transport of Gaussian mixtures."""

from typing import Callable, Iterable, List, Optional, Tuple

import jax
import jax.numpy as jnp


def get_mean_and_var(
    points: jnp.ndarray,  # (n, d)
    weights: jnp.ndarray,  # (n,)
) -> Tuple[jnp.ndarray, jnp.ndarray]:
  """Get the mean and variance of a weighted set of points."""
  weights_sum = jnp.sum(weights, axis=-1)  # (1,)
  mean = (
      # matmul((1, n), (n, d)) -> (1, d)
      jnp.matmul(weights, points) / weights_sum
  )
  # center points
  centered = points - mean[None, :]  # (n, d) - (1, d)
  var = (
      # matmul((1, n), (n, d)) -> (1, d)
      jnp.matmul(weights, centered ** 2.) / weights_sum
  )
  return mean, var


def get_mean_and_cov(
    points: jnp.ndarray,  # (n, d)
    weights: jnp.ndarray,  # (n,)
) -> Tuple[jnp.ndarray, jnp.ndarray]:
  """Get the mean and covariance of a weighted set of points."""
  weights_sum = jnp.sum(weights, axis=-1, keepdims=True)  # (1,)
  mean = (
      # matmul((1, n), (n, d)) -> (1, d)
      jnp.matmul(weights, points) / weights_sum
  )
  # center points
  centered = points - mean[None, :]  # (n, d) - (1, d)
  cov = (
      jnp.matmul(
          # (1, n)           (d, n)
          weights[None, :] * jnp.swapaxes(centered, axis1=-2, axis2=-1),
          # (n, d)
          centered
      ) / weights_sum
  )
  return mean, cov


def flat_to_tril(x: jnp.ndarray, size: int) -> jnp.ndarray:
  """Map flat values to lower triangular matrices.

  Args:
    x: flat values
    size: size of lower triangular matrices. x should have shape
      (..., size(size+1)/2), and the final matrices should have shape
      (..., size, size).

  Returns:
    Lower triangular matrices.
  """
  m = jnp.zeros(tuple(list(x.shape[:-1]) + [size, size]), dtype=x.dtype)
  tril = jnp.tril_indices(size)
  return m.at[..., tril[0], tril[1]].set(x)


def tril_to_flat(m: jnp.ndarray) -> jnp.ndarray:
  """Flatten lower triangular matrices.

  Args:
    m: lower triangular matrices of shape (..., size, size)

  Returns:
    A vector of shape (..., size (size+1) // 2)
  """
  size = m.shape[-1]
  tril = jnp.tril_indices(size)
  return m[..., tril[0], tril[1]]


def apply_to_diag(
    m: jnp.ndarray, fn: Callable[[jnp.ndarray], jnp.ndarray]
) -> jnp.ndarray:
  """Apply a function to the diagonal of a matrix."""
  size = m.shape[-1]
  diag = jnp.diagonal(m, axis1=-2, axis2=-1)
  ind = jnp.arange(size, dtype=jnp.int32)
  return m.at[..., ind, ind].set(fn(diag))


def matrix_powers(
    m: jnp.ndarray,
    powers: Iterable[float],
) -> List[jnp.ndarray]:
  """Raise a real, symmetric matrix to multiple powers."""
  eigs, q = jnp.linalg.eigh(m)
  qt = jnp.swapaxes(q, axis1=-2, axis2=-1)
  ret = []
  for power in powers:
    ret.append(jnp.matmul(jnp.expand_dims(eigs ** power, -2) * q, qt))
  return ret


def invmatvectril(
    m: jnp.ndarray, x: jnp.ndarray, lower: bool = True
) -> jnp.ndarray:
  """Multiply x by the inverse of a triangular matrix.

  Args:
    m: triangular matrix, shape (d, d)
    x: array of points, shape (n, d)
    lower: if True, m is lower triangular; otherwise m is upper triangular

  Returns:
    m^{-1} x
  """
  return jnp.transpose(
      jax.scipy.linalg.solve_triangular(m, jnp.transpose(x), lower=lower)
  )


def get_random_orthogonal(
    key: jnp.ndarray,
    dim: int,
    dtype: Optional[jnp.dtype] = None
) -> jnp.ndarray:
  """Get a random orthogonal matrix with the specified dimension."""
  m = jax.random.normal(key=key, shape=[dim, dim], dtype=dtype)
  q, _ = jnp.linalg.qr(m)
  return q



# Copyright 2022 The OTT Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Sinkhorn initializers."""
import abc
from typing import TYPE_CHECKING, Any, Dict, Optional, Sequence, Tuple

import jax
import jax.numpy as jnp

from ott.geometry import pointcloud

if TYPE_CHECKING:
  from ott.problems.linear import linear_problem

__all__ = ["DefaultInitializer", "GaussianInitializer", "SortingInitializer"]


@jax.tree_util.register_pytree_node_class
class SinkhornInitializer(abc.ABC):
  """Base class for Sinkhorn initializers."""

  @abc.abstractmethod
  def init_dual_a(
      self, ot_prob: 'linear_problem.LinearProblem', lse_mode: bool
  ) -> jnp.ndarray:
    """Initialization for Sinkhorn potential/scaling f_u."""

  @abc.abstractmethod
  def init_dual_b(
      self, ot_prob: 'linear_problem.LinearProblem', lse_mode: bool
  ) -> jnp.ndarray:
    """Initialization for Sinkhorn potential/scaling g_v."""

  def __call__(
      self,
      ot_prob: 'linear_problem.LinearProblem',
      a: Optional[jnp.ndarray],
      b: Optional[jnp.ndarray],
      lse_mode: bool,
  ) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Initialize Sinkhorn potentials/scalings f_u and g_v.

    Args:
      ot_prob: Linear OT problem.
      a: Initial potential/scaling f_u. If ``None``, it will be initialized using
        :meth:`init_dual_a`.
      b: Initial potential/scaling g_v. If ``None``, it will be initialized using
        :meth:`init_dual_b`.
      lse_mode: Return potentials if true, scalings otherwise.

    Returns:
      The initial potentials/scalings.
    """
    n, m = ot_prob.geom.shape
    if a is None:
      a = self.init_dual_a(ot_prob, lse_mode=lse_mode)
    if b is None:
      b = self.init_dual_b(ot_prob, lse_mode=lse_mode)

    assert a.shape == (
        n,
    ), f"Expected `f_u` to have shape `{n,}`, found `{a.shape}`."
    assert b.shape == (
        m,
    ), f"Expected `g_v` to have shape `{m,}`, found `{b.shape}`."

    # cancel dual variables for zero weights
    a = jnp.where(ot_prob.a > 0., a, -jnp.inf if lse_mode else 0.)
    b = jnp.where(ot_prob.b > 0., b, -jnp.inf if lse_mode else 0.)

    return a, b

  def tree_flatten(self) -> Tuple[Sequence[Any], Dict[str, Any]]:
    return [], {}

  @classmethod
  def tree_unflatten(
      cls, aux_data: Dict[str, Any], children: Sequence[Any]
  ) -> "SinkhornInitializer":
    return cls(*children, **aux_data)


@jax.tree_util.register_pytree_node_class
class DefaultInitializer(SinkhornInitializer):
  """Default initialization of Sinkhorn dual potentials/primal scalings."""

  def init_dual_a(
      self, ot_prob: 'linear_problem.LinearProblem', lse_mode: bool
  ) -> jnp.ndarray:
    """Initialize Sinkhorn potential/scaling f_u.

    Args:
      ot_prob: OT problem between discrete distributions of size n and m.
      lse_mode: Return potential if true, scaling if false.

    Returns:
      potential/scaling, array of size n.
    """
    a = ot_prob.a
    init_dual_a = jnp.zeros_like(a) if lse_mode else jnp.ones_like(a)
    return init_dual_a

  def init_dual_b(
      self, ot_prob: 'linear_problem.LinearProblem', lse_mode: bool
  ) -> jnp.ndarray:
    """Initialize Sinkhorn potential/scaling g_v.

    Args:
      ot_prob: OT problem between discrete distributions of size n and m.
      lse_mode: Return potential if true, scaling if false.

    Returns:
      potential/scaling, array of size m.
    """
    b = ot_prob.b
    init_dual_b = jnp.zeros_like(b) if lse_mode else jnp.ones_like(b)
    return init_dual_b


@jax.tree_util.register_pytree_node_class
class GaussianInitializer(DefaultInitializer):
  """Gaussian initializer :cite:`thornton2022rethinking:22`.

  Compute Gaussian approximations of each point cloud, then compute closed from
  Kantorovich potential between Gaussian approximations using Brenier's theorem
  (adapt convex/Brenier potential to Kantorovich). Use this Gaussian potential
  to initialize Sinkhorn potentials/scalings.
  """

  def init_dual_a(
      self,
      ot_prob: 'linear_problem.LinearProblem',
      lse_mode: bool,
  ) -> jnp.ndarray:
    """Gaussian initialization function.

    Args:
      ot_prob: OT problem between discrete distributions of size n and m.
      lse_mode: Return potential if true, scaling if false.

    Returns:
      potential/scaling, array of size n.
    """
    # import Gaussian here due to circular imports
    from ott.tools.gaussian_mixture import gaussian

    assert isinstance(
        ot_prob.geom, pointcloud.PointCloud
    ), "Gaussian initializer valid only for point clouds."

    x, y = ot_prob.geom.x, ot_prob.geom.y
    a, b = ot_prob.a, ot_prob.b

    gaussian_a = gaussian.Gaussian.from_samples(x, weights=a)
    gaussian_b = gaussian.Gaussian.from_samples(y, weights=b)
    # Brenier potential for cost ||x-y||^2/2, multiply by two for ||x-y||^2
    f_potential = 2 * gaussian_a.f_potential(dest=gaussian_b, points=x)
    f_potential = f_potential - jnp.mean(f_potential)
    f_u = f_potential if lse_mode else ot_prob.geom.scaling_from_potential(
        f_potential
    )
    return f_u


@jax.tree_util.register_pytree_node_class
class SortingInitializer(DefaultInitializer):
  """Sorting initializer :cite:`thornton2022rethinking:22`.

  Solves non-regularized OT problem via sorting, then compute potential through
  iterated minimum on C-transform and use this potential to initialize
  regularized potential.

  Args:
    vectorized_update: Use vectorized inner loop if true.
    tolerance: DualSort convergence threshold.
    max_iter: Max DualSort steps.
  """

  def __init__(
      self,
      vectorized_update: bool = True,
      tolerance: float = 1e-2,
      max_iter: int = 100
  ):
    super().__init__()
    self.tolerance = tolerance
    self.max_iter = max_iter
    self.vectorized_update = vectorized_update

  def tree_flatten(self) -> Tuple[Sequence[Any], Dict[str, Any]]:
    return ([], {
        'tolerance': self.tolerance,
        'max_iter': self.max_iter,
        'vectorized_update': self.vectorized_update
    })

  def _init_sorting_dual(
      self, modified_cost: jnp.ndarray, init_f: jnp.ndarray
  ) -> jnp.ndarray:
    """Run DualSort algorithm.

    Args:
      modified_cost: cost matrix minus diagonal column-wise.
      init_f: potential f, array of size n. This is the starting potential,
        which is then updated to make the init potential, so an init of an init.

    Returns:
      potential f, array of size n.
    """

    def body_fn(
        state: Tuple[jnp.ndarray, float, int]
    ) -> Tuple[jnp.ndarray, float, int]:
      prev_f, _, it = state
      new_f = fn(prev_f, modified_cost)
      diff = jnp.sum((new_f - prev_f) ** 2)
      it += 1
      return new_f, diff, it

    def cond_fn(state: Tuple[jnp.ndarray, float, int]) -> bool:
      _, diff, it = state
      return jnp.logical_and(diff > self.tolerance, it < self.max_iter)

    fn = _vectorized_update if self.vectorized_update else _coordinate_update
    state = (init_f, jnp.inf, 0)  # init, error, iter
    f_potential, _, _ = jax.lax.while_loop(
        cond_fun=cond_fn, body_fun=body_fn, init_val=state
    )

    return f_potential

  def init_dual_a(
      self,
      ot_prob: 'linear_problem.LinearProblem',
      lse_mode: bool,
      init_f: Optional[jnp.ndarray] = None,
  ) -> jnp.ndarray:
    """Apply DualSort algorithm.

    Args:
      ot_prob: OT problem.
      lse_mode: Return potential if true, scaling if false.
      init_f: potential f, array of size n. This is the starting potential,
        which is then updated to make the init potential, so an init of an init.

    Returns:
      potential/scaling f_u, array of size n.
    """
    assert not ot_prob.geom.is_online, \
        "Sorting initializer does not work for online geometry."
    # check for sorted x, y requires point cloud and could slow initializer
    cost_matrix = ot_prob.geom.cost_matrix

    assert cost_matrix.shape[0] == cost_matrix.shape[
        1], "Requires square cost matrix."

    modified_cost = cost_matrix - jnp.diag(cost_matrix)[None, :]

    n = cost_matrix.shape[0]
    init_f = jnp.zeros(n) if init_f is None else init_f

    f_potential = self._init_sorting_dual(modified_cost, init_f)
    f_potential = f_potential - jnp.mean(f_potential)

    f_u = f_potential if lse_mode else ot_prob.geom.scaling_from_potential(
        f_potential
    )

    return f_u


def _vectorized_update(
    f: jnp.ndarray, modified_cost: jnp.ndarray
) -> jnp.ndarray:
  """Inner loop DualSort Update.

  Args:
    f : potential f, array of size n.
    modified_cost: cost matrix minus diagonal column-wise.

  Returns:
    updated potential vector, f.
  """
  return jnp.min(modified_cost + f[None, :], axis=1)


def _coordinate_update(
    f: jnp.ndarray, modified_cost: jnp.ndarray
) -> jnp.ndarray:
  """Coordinate-wise updates within inner loop.

  Args:
    f: potential f, array of size n.
    modified_cost: cost matrix minus diagonal column-wise.

  Returns:
    updated potential vector, f.
  """

  def body_fn(i: int, f: jnp.ndarray) -> jnp.ndarray:
    new_f = jnp.min(modified_cost[i, :] + f)
    return f.at[i].set(new_f)

  return jax.lax.fori_loop(0, len(f), body_fn, f)



# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""A Jax implementation of the Low-Rank Sinkhorn algorithm."""
from typing import Any, Mapping, NamedTuple, NoReturn, Optional, Tuple, Union

import jax
import jax.numpy as jnp
import jax.scipy as jsp
from typing_extensions import Literal

from ott.geometry import geometry, low_rank, pointcloud
from ott.initializers.linear import initializers_lr as init_lib
from ott.math import fixed_point_loop
from ott.math import utils as mu
from ott.problems.linear import linear_problem
from ott.solvers.linear import sinkhorn

__all__ = ["LRSinkhorn", "LRSinkhornOutput"]


class LRSinkhornState(NamedTuple):
  """State of the Low Rank Sinkhorn algorithm."""

  q: jnp.ndarray
  r: jnp.ndarray
  g: jnp.ndarray
  gamma: float
  costs: jnp.ndarray
  errors: jnp.ndarray
  crossed_threshold: bool

  def compute_error(self, previous_state: "LRSinkhornState") -> float:
    err_1 = mu.js(self.q, previous_state.q, c=1.)
    err_2 = mu.js(self.r, previous_state.r, c=1.)
    err_3 = mu.js(self.g, previous_state.g, c=1.)

    return ((1. / self.gamma) ** 2) * (err_1 + err_2 + err_3)

  def reg_ot_cost(
      self,
      ot_prob: linear_problem.LinearProblem,
      use_danskin: bool = False
  ) -> float:
    return compute_reg_ot_cost(self.q, self.r, self.g, ot_prob, use_danskin)

  def solution_error(
      self, ot_prob: linear_problem.LinearProblem, norm_error: Tuple[int, ...],
      lse_mode: bool
  ) -> jnp.ndarray:
    return solution_error(self.q, self.r, ot_prob, norm_error, lse_mode)

  def set(self, **kwargs: Any) -> 'LRSinkhornState':
    """Return a copy of self, with potential overwrites."""
    return self._replace(**kwargs)


def compute_reg_ot_cost(
    q: jnp.ndarray,
    r: jnp.ndarray,
    g: jnp.ndarray,
    ot_prob: linear_problem.LinearProblem,
    use_danskin: bool = False
) -> float:
  q = jax.lax.stop_gradient(q) if use_danskin else q
  r = jax.lax.stop_gradient(r) if use_danskin else r
  g = jax.lax.stop_gradient(g) if use_danskin else g
  return jnp.sum(ot_prob.geom.apply_cost(r, axis=1) * q * (1. / g)[None, :])


def solution_error(
    q: jnp.ndarray, r: jnp.ndarray, ot_prob: linear_problem.LinearProblem,
    norm_error: Tuple[int, ...], lse_mode: bool
) -> jnp.ndarray:
  """Compute solution error.

  Since only balanced case is available for LR, this is marginal deviation.

  Args:
    q: first factor of solution
    r: second factor of solution
    ot_prob: linear problem
    norm_error: int, p-norm used to compute error.
    lse_mode: True if log-sum-exp operations, False if kernel vector products.

  Returns:
    one or possibly many numbers quantifying deviation to true marginals.
  """
  del lse_mode
  norm_error = jnp.array(norm_error)
  # Update the error
  err = jnp.sum(
      jnp.abs(jnp.sum(q, axis=1) - ot_prob.a) ** norm_error[:, None], axis=1
  ) ** (1.0 / norm_error)
  err += jnp.sum(
      jnp.abs(jnp.sum(r, axis=1) - ot_prob.b) ** norm_error[:, None], axis=1
  ) ** (1.0 / norm_error)
  err += jnp.sum(
      jnp.abs(jnp.sum(q, axis=0) - jnp.sum(r, axis=0)) ** norm_error[:, None],
      axis=1
  ) ** (1.0 / norm_error)

  return err


class LRSinkhornOutput(NamedTuple):
  """Implement the problems.Transport interface, for a LR Sinkhorn solution."""

  q: jnp.ndarray
  r: jnp.ndarray
  g: jnp.ndarray
  costs: jnp.ndarray
  # TODO(michalk8): must be called `errors`, because of `store_inner_errors`
  # in future, enforce via class hierarchy
  errors: jnp.ndarray
  ot_prob: linear_problem.LinearProblem
  # TODO(michalk8): Optional is an artifact of the current impl., refactor
  reg_ot_cost: Optional[float] = None

  def set(self, **kwargs: Any) -> 'LRSinkhornOutput':
    """Return a copy of self, with potential overwrites."""
    return self._replace(**kwargs)

  def set_cost(
      self,
      ot_prob: linear_problem.LinearProblem,
      lse_mode: bool,
      use_danskin: bool = False
  ) -> 'LRSinkhornOutput':
    del lse_mode
    return self.set(reg_ot_cost=self.compute_reg_ot_cost(ot_prob, use_danskin))

  def compute_reg_ot_cost(
      self,
      ot_prob: linear_problem.LinearProblem,
      use_danskin: bool = False,
  ) -> float:
    return compute_reg_ot_cost(self.q, self.r, self.g, ot_prob, use_danskin)

  @property
  def linear(self) -> bool:
    return isinstance(self.ot_prob, linear_problem.LinearProblem)

  @property
  def geom(self) -> geometry.Geometry:
    return self.ot_prob.geom

  @property
  def a(self) -> jnp.ndarray:
    return self.ot_prob.a

  @property
  def b(self) -> jnp.ndarray:
    return self.ot_prob.b

  @property
  def linear_output(self) -> bool:
    return True

  @property
  def converged(self) -> bool:
    return jnp.logical_and(
        jnp.any(self.costs == -1), jnp.all(jnp.isfinite(self.costs))
    )

  @property
  def matrix(self) -> jnp.ndarray:
    """Transport matrix if it can be instantiated."""
    return (self.q * self._inv_g) @ self.r.T

  def apply(self, inputs: jnp.ndarray, axis: int = 0) -> jnp.ndarray:
    """Apply the transport to a array; axis=1 for its transpose."""
    q, r = (self.q, self.r) if axis == 1 else (self.r, self.q)
    # for `axis=0`: (batch, m), (m, r), (r,), (r, n)
    return ((inputs @ r) * self._inv_g) @ q.T

  def marginal(self, axis: int) -> jnp.ndarray:
    length = self.q.shape[0] if axis == 0 else self.r.shape[0]
    return self.apply(jnp.ones(length,), axis=axis)

  def cost_at_geom(self, other_geom: geometry.Geometry) -> float:
    """Return OT cost for matrix, evaluated at other cost matrix."""
    return jnp.sum(self.q * other_geom.apply_cost(self.r, axis=1) * self._inv_g)

  # TODO(michalk8): when refactoring the API, use a property
  def transport_mass(self) -> float:
    """Sum of transport matrix."""
    return self.marginal(0).sum()

  @property
  def _inv_g(self) -> jnp.ndarray:
    return 1. / self.g


@jax.tree_util.register_pytree_node_class
class LRSinkhorn(sinkhorn.Sinkhorn):
  r"""A Low-Rank Sinkhorn solver for linear reg-OT problems.

  The algorithm is described in :cite:`scetbon:21` and the implementation
  contained here is adapted from `LOT <https://github.com/meyerscetbon/LOT>`_.

  The algorithm minimizes a non-convex problem. It therefore requires special
  care to initialization and convergence. Convergence is evaluated on successive
  evaluations of the objective. The algorithm is only provided for the balanced
  case.

  Args:
    rank: The rank constraint on the coupling to minimize the linear OT problem
    gamma: The (inverse of) gradient step size used by mirror descent.
    gamma_rescale: Whether to rescale :math:`\gamma` every iteration as
      described in :cite:`scetbon:22b`.
    epsilon: Entropic regularization added on top of low-rank problem.
    initializer: How to initialize the :math:`Q`, :math:`R` and :math:`g`
      factors. Valid options are:

        - `'random'` - :class:`~ott.initializers.linear.initializers_lr.RandomInitializer`.
        - `'rank2'` - :class:`~ott.initializers.linear.initializers_lr.Rank2Initializer`.
        - `'k-means'` - :class:`~ott.initializers.linear.initializers_lr.KMeansInitializer`.
        - `'generalized-k-means'` - :class:`~ott.initializers.linear.initializers_lr.GeneralizedKMeansInitializer`.

      If `None`, :class:`~ott.initializers.linear.initializers_lr.KMeansInitializer`
      is used when the linear problem's geometry is
      :class:`~ott.geometry.pointcloud.PointCloud` or
      :class:`~ott.geometry.low_rank.LRCGeometry`.
      Otherwise, use :class:`~ott.initializers.linear.initializers_lr.RandomInitializer`.

    lse_mode: Whether to run computations in lse or kernel mode. At the moment,
      only ``lse_mode = True`` is implemented.
    inner_iterations: Number of inner iterations used by the algorithm before
      re-evaluating progress.
    use_danskin: Use Danskin theorem to evaluate gradient of objective w.r.t.
      input parameters. Only `True` handled at this moment.
    implicit_diff: Whether to use implicit differentiation. Currently, only
      ``implicit_diff = False`` is implemented.
    kwargs_dys: Keyword arguments passed to :meth:`dykstra_update`.
    kwargs_init: Keyword arguments for
      :class:`~ott.initializers.linear.initializers_lr.LRInitializer`.
    kwargs: Keyword arguments for
      :class:`~ott.solvers.linear.sinkhorn.Sinkhorn`.
  """

  def __init__(
      self,
      rank: int,
      gamma: float = 10.,
      gamma_rescale: bool = True,
      epsilon: float = 0.,
      initializer: Optional[Union[Literal["random", "rank2", "k-means",
                                          "generalized-k-means"],
                                  init_lib.LRInitializer]] = "random",
      lse_mode: bool = True,
      inner_iterations: int = 10,
      use_danskin: bool = True,
      implicit_diff: bool = False,
      kwargs_dys: Optional[Mapping[str, Any]] = None,
      kwargs_init: Optional[Mapping[str, Any]] = None,
      **kwargs: Any,
  ):
    assert lse_mode, "Kernel mode not yet implemented."
    assert not implicit_diff, "Implicit diff. not yet implemented."
    super().__init__(
        lse_mode=lse_mode,
        inner_iterations=inner_iterations,
        use_danskin=use_danskin,
        implicit_diff=implicit_diff,
        **kwargs
    )
    self.rank = rank
    self.gamma = gamma
    self.gamma_rescale = gamma_rescale
    self.epsilon = epsilon
    self.initializer = initializer
    # can be `None`
    self.kwargs_dys = {} if kwargs_dys is None else kwargs_dys
    self.kwargs_init = {} if kwargs_init is None else kwargs_init

  def __call__(
      self,
      ot_prob: linear_problem.LinearProblem,
      init: Tuple[Optional[jnp.ndarray], Optional[jnp.ndarray],
                  Optional[jnp.ndarray]] = (None, None, None),
      key: Optional[jnp.ndarray] = None,
      **kwargs: Any,
  ) -> LRSinkhornOutput:
    """Run low-rank Sinkhorn.

    Args:
      ot_prob: Linear OT problem.
      init: Initial values for the low-rank factors:

        - :attr:`~ott.solvers.linear.sinkhorn_lr.LRSinkhornOutput.q`.
        - :attr:`~ott.solvers.linear.sinkhorn_lr.LRSinkhornOutput.r`.
        - :attr:`~ott.solvers.linear.sinkhorn_lr.LRSinkhornOutput.g`.

        Any `None` values will be initialized using the initializer.
      key: Random key for seeding.
      kwargs: Additional arguments when calling the initializer.

    Returns:
      The low-rank Sinkhorn output.
    """
    assert ot_prob.is_balanced, "Unbalanced case is not implemented."
    initializer = self.create_initializer(ot_prob)
    init = initializer(ot_prob, *init, key=key, **kwargs)
    run_fn = jax.jit(run) if self.jit else run
    return run_fn(ot_prob, self, init)

  def _lr_costs(
      self,
      ot_prob: linear_problem.LinearProblem,
      state: LRSinkhornState,
  ) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, float]:
    log_q, log_r, log_g = (
        mu.safe_log(state.q), mu.safe_log(state.r), mu.safe_log(state.g)
    )

    grad_q = ot_prob.geom.apply_cost(state.r, axis=1) / state.g[None, :]
    grad_r = ot_prob.geom.apply_cost(state.q) / state.g[None, :]
    diag_qcr = jnp.sum(
        state.q * ot_prob.geom.apply_cost(state.r, axis=1), axis=0
    )
    grad_g = -diag_qcr / (state.g ** 2)
    if self.is_entropic:
      grad_q += self.epsilon * log_q
      grad_r += self.epsilon * log_r
      grad_g += self.epsilon * log_g

    if self.gamma_rescale:
      norm_q = jnp.max(jnp.abs(grad_q)) ** 2
      norm_r = jnp.max(jnp.abs(grad_r)) ** 2
      norm_g = jnp.max(jnp.abs(grad_g)) ** 2
      gamma = self.gamma / jnp.max(jnp.array([norm_q, norm_r, norm_g]))
    else:
      gamma = self.gamma

    c_q = grad_q - (1. / gamma) * log_q
    c_r = grad_r - (1. / gamma) * log_r
    h = -grad_g + (1. / gamma) * log_g
    return c_q, c_r, h, gamma

  def dykstra_update(
      self,
      c_q: jnp.ndarray,
      c_r: jnp.ndarray,
      h: jnp.ndarray,
      gamma: float,
      ot_prob: linear_problem.LinearProblem,
      min_entry_value: float = 1e-6,
      tolerance: float = 1e-3,
      min_iter: int = 0,
      inner_iter: int = 10,
      max_iter: int = 10000
  ) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    # shortcuts for problem's definition.
    r = self.rank
    n, m = ot_prob.geom.shape
    loga, logb = jnp.log(ot_prob.a), jnp.log(ot_prob.b)

    h_old = h
    g1_old, g2_old = jnp.zeros(r), jnp.zeros(r)
    f1, f2 = jnp.zeros(n), jnp.zeros(m)

    w_gi, w_gp = jnp.zeros(r), jnp.zeros(r)
    w_q, w_r = jnp.zeros(r), jnp.zeros(r)
    err = jnp.inf
    state_inner = f1, f2, g1_old, g2_old, h_old, w_gi, w_gp, w_q, w_r, err
    constants = c_q, c_r, loga, logb

    def cond_fn(
        iteration: int, constants: Tuple[jnp.ndarray, ...],
        state_inner: Tuple[jnp.ndarray, ...]
    ) -> bool:
      del iteration, constants
      *_, err = state_inner
      return err > tolerance

    def _softm(
        f: jnp.ndarray, g: jnp.ndarray, c: jnp.ndarray, axis: int
    ) -> jnp.ndarray:
      return jsp.special.logsumexp(
          gamma * (f[:, None] + g[None, :] - c), axis=axis
      )

    def body_fn(
        iteration: int, constants: Tuple[jnp.ndarray, ...],
        state_inner: Tuple[jnp.ndarray, ...], compute_error: bool
    ) -> Tuple[jnp.ndarray, ...]:
      # TODO(michalk8): in the future, use `NamedTuple`
      f1, f2, g1_old, g2_old, h_old, w_gi, w_gp, w_q, w_r, err = state_inner
      c_q, c_r, loga, logb = constants

      # First Projection
      f1 = jnp.where(
          jnp.isfinite(loga),
          (loga - _softm(f1, g1_old, c_q, axis=1)) / gamma + f1, loga
      )
      f2 = jnp.where(
          jnp.isfinite(logb),
          (logb - _softm(f2, g2_old, c_r, axis=1)) / gamma + f2, logb
      )

      h = h_old + w_gi
      h = jnp.maximum(jnp.log(min_entry_value) / gamma, h)
      w_gi += h_old - h
      h_old = h

      # Update couplings
      g_q = _softm(f1, g1_old, c_q, axis=0)
      g_r = _softm(f2, g2_old, c_r, axis=0)

      # Second Projection
      h = (1. / 3.) * (h_old + w_gp + w_q + w_r)
      h += g_q / (3. * gamma)
      h += g_r / (3. * gamma)
      g1 = h + g1_old - g_q / gamma
      g2 = h + g2_old - g_r / gamma

      w_q = w_q + g1_old - g1
      w_r = w_r + g2_old - g2
      w_gp = h_old + w_gp - h

      q, r, _ = recompute_couplings(f1, g1, c_q, f2, g2, c_r, h, gamma)

      g1_old = g1
      g2_old = g2
      h_old = h

      err = jax.lax.cond(
          jnp.logical_and(compute_error, iteration >= min_iter),
          lambda: solution_error(q, r, ot_prob, self.norm_error, self.lse_mode)[
              0], lambda: err
      )

      return f1, f2, g1_old, g2_old, h_old, w_gi, w_gp, w_q, w_r, err

    def recompute_couplings(
        f1: jnp.ndarray,
        g1: jnp.ndarray,
        c_q: jnp.ndarray,
        f2: jnp.ndarray,
        g2: jnp.ndarray,
        c_r: jnp.ndarray,
        h: jnp.ndarray,
        gamma: float,
    ) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
      q = jnp.exp(gamma * (f1[:, None] + g1[None, :] - c_q))
      r = jnp.exp(gamma * (f2[:, None] + g2[None, :] - c_r))
      g = jnp.exp(gamma * h)
      return q, r, g

    state_inner = fixed_point_loop.fixpoint_iter_backprop(
        cond_fn, body_fn, min_iter, max_iter, inner_iter, constants, state_inner
    )

    f1, f2, g1_old, g2_old, h_old, _, _, _, _, _ = state_inner
    return recompute_couplings(f1, g1_old, c_q, f2, g2_old, c_r, h_old, gamma)

  def lse_step(
      self, ot_prob: linear_problem.LinearProblem, state: LRSinkhornState,
      iteration: int
  ) -> LRSinkhornState:
    """LR Sinkhorn LSE update."""
    c_q, c_r, h, gamma = self._lr_costs(ot_prob, state)
    q, r, g = self.dykstra_update(
        c_q, c_r, h, gamma, ot_prob, **self.kwargs_dys
    )
    return state.set(q=q, g=g, r=r, gamma=gamma)

  def kernel_step(
      self, ot_prob: linear_problem.LinearProblem, state: LRSinkhornState,
      iteration: int
  ) -> NoReturn:
    """Not implemented."""
    # TODO(cuturi): kernel step not implemented.
    raise NotImplementedError("Not implemented.")

  def one_iteration(
      self, ot_prob: linear_problem.LinearProblem, state: LRSinkhornState,
      iteration: int, compute_error: bool
  ) -> LRSinkhornState:
    """Carries out one LR sinkhorn iteration.

    Depending on lse_mode, these iterations can be either in:

      - log-space for numerical stability.
      - scaling space, using standard kernel-vector multiply operations.

    Args:
      ot_prob: the transport problem definition
      state: LRSinkhornState named tuple.
      iteration: the current iteration of the Sinkhorn outer loop.
      compute_error: flag to indicate this iteration computes/stores an error

    Returns:
      The updated state.
    """
    previous_state = state
    it = iteration // self.inner_iterations
    if self.lse_mode:  # In lse_mode, run additive updates.
      state = self.lse_step(ot_prob, state, iteration)
    else:
      state = self.kernel_step(ot_prob, state, iteration)

    # re-computes error if compute_error is True, else set it to inf.
    cost = jax.lax.cond(
        jnp.logical_and(compute_error, iteration >= self.min_iterations),
        lambda: state.reg_ot_cost(ot_prob), lambda: jnp.inf
    )
    error = state.compute_error(previous_state)
    crossed_threshold = jnp.logical_or(
        state.crossed_threshold,
        jnp.logical_and(
            state.errors[it - 1] >= self.threshold, error < self.threshold
        )
    )

    return state.set(
        costs=state.costs.at[it].set(cost),
        errors=state.errors.at[it].set(error),
        crossed_threshold=crossed_threshold,
    )

  @property
  def norm_error(self) -> Tuple[int]:
    return self._norm_error,

  @property
  def is_entropic(self) -> bool:
    """Whether entropy regularization is used."""
    return self.epsilon > 0.

  def create_initializer(
      self, prob: linear_problem.LinearProblem
  ) -> init_lib.LRInitializer:
    """Create a low-rank Sinkhorn initializer.

    Args:
      prob: Linear OT problem used to determine the initializer.

    Returns:
      Low-rank initializer.
    """
    if isinstance(self.initializer, init_lib.LRInitializer):
      initializer = self.initializer
    elif self.initializer is None:
      kind = "k-means" if isinstance(
          prob.geom, (pointcloud.PointCloud, low_rank.LRCGeometry)
      ) else "random"
      initializer = init_lib.LRInitializer.from_solver(
          self, kind=kind, **self.kwargs_init
      )
    else:
      initializer = init_lib.LRInitializer.from_solver(
          self, kind=self.initializer, **self.kwargs_init
      )

    assert initializer.rank == self.rank, \
        f"Expected initializer of rank `{self.rank}`, " \
        f"found `{initializer.rank}`."
    return initializer

  def init_state(
      self, ot_prob: linear_problem.LinearProblem,
      init: Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]
  ) -> LRSinkhornState:
    """Return the initial state of the loop."""
    q, r, g = init
    return LRSinkhornState(
        q=q,
        r=r,
        g=g,
        gamma=self.gamma,
        costs=-jnp.ones(self.outer_iterations),
        errors=-jnp.ones(self.outer_iterations),
        crossed_threshold=False,
    )

  def output_from_state(
      self, ot_prob: linear_problem.LinearProblem, state: LRSinkhornState
  ) -> LRSinkhornOutput:
    """Create an output from a loop state.

    Args:
      ot_prob: the transport problem.
      state: a LRSinkhornState.

    Returns:
      A LRSinkhornOutput.
    """
    return LRSinkhornOutput(
        q=state.q,
        r=state.r,
        g=state.g,
        ot_prob=ot_prob,
        costs=state.costs,
        errors=state.errors,
    )

  def _converged(self, state: LRSinkhornState, iteration: int) -> bool:

    def conv_crossed(prev_err: float, curr_err: float) -> bool:
      return jnp.logical_and(
          prev_err < self.threshold, curr_err < self.threshold
      )

    def conv_not_crossed(prev_err: float, curr_err: float) -> bool:
      return jnp.logical_and(curr_err < prev_err, curr_err < self.threshold)

    # for convergence error, we consider 2 possibilities:
    # 1. we either crossed the convergence threshold; in this case we require
    #   that the previous error was also below the threshold
    # 2. we haven't crossed the threshold; in this case, we can be below or
    #   above the threshold:
    #     if we're above, we wait until we reach the convergence threshold and
    #     then, the above condition applies
    #     if we're below and we improved w.r.t. the previous iteration,
    #     we have converged; otherwise we continue, since we may be stuck
    #     in a local minimum (e.g., during the initial iterations)

    it = iteration // self.inner_iterations
    return jax.lax.cond(
        state.crossed_threshold, conv_crossed, conv_not_crossed,
        state.errors[it - 2], state.errors[it - 1]
    )

  def _diverged(self, state: LRSinkhornState, iteration: int) -> bool:
    it = iteration // self.inner_iterations
    return jnp.logical_and(
        jnp.logical_not(jnp.isfinite(state.errors[it - 1])),
        jnp.logical_not(jnp.isfinite(state.costs[it - 1]))
    )


def run(
    ot_prob: linear_problem.LinearProblem,
    solver: LRSinkhorn,
    init: Tuple[Optional[jnp.ndarray], Optional[jnp.ndarray],
                Optional[jnp.ndarray]],
) -> LRSinkhornOutput:
  """Run loop of the solver, outputting a state upgraded to an output."""
  out = sinkhorn.iterations(ot_prob, solver, init)
  out = out.set_cost(
      ot_prob, lse_mode=solver.lse_mode, use_danskin=solver.use_danskin
  )
  return out.set(ot_prob=ot_prob)


def make(
    rank: int,
    gamma: float = 1.0,
    epsilon: float = 1e-4,
    initializer: Literal['random', 'rank2', 'k-means'] = 'k-means',
    lse_mode: bool = True,
    threshold: float = 1e-3,
    norm_error: int = 10,
    inner_iterations: int = 1,
    min_iterations: int = 0,
    max_iterations: int = 2000,
    use_danskin: bool = True,
    implicit_diff: bool = False,
    jit: bool = True,
    kwargs_dys: Optional[Mapping[str, Any]] = None
) -> LRSinkhorn:
  return LRSinkhorn(
      rank=rank,
      gamma=gamma,
      epsilon=epsilon,
      initializer=initializer,
      lse_mode=lse_mode,
      threshold=threshold,
      norm_error=norm_error,
      inner_iterations=inner_iterations,
      min_iterations=min_iterations,
      max_iterations=max_iterations,
      use_danskin=use_danskin,
      implicit_diff=implicit_diff,
      jit=jit,
      kwargs_dys=kwargs_dys
  )
